import copy
from unittest.mock import MagicMock

import numpy as np
import pytest
import torch
import torch.nn as nn
import torch.optim as optim
from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer
from gymnasium import spaces
from gymnasium.spaces import Box, Discrete
from pettingzoo import ParallelEnv
from torch._dynamo import OptimizedModule

from agilerl.algorithms.maddpg import MADDPG
from agilerl.modules import EvolvableCNN, EvolvableMLP, EvolvableMultiInput, ModuleDict
from agilerl.modules.custom_components import GumbelSoftmax
from agilerl.networks.actors import DeterministicActor
from agilerl.networks.q_networks import ContinuousQNetwork
from agilerl.utils.algo_utils import concatenate_spaces
from agilerl.utils.evolvable_networks import get_default_encoder_config
from agilerl.utils.utils import make_multi_agent_vect_envs
from agilerl.wrappers.make_evolvable import MakeEvolvable
from tests.helper_functions import assert_not_equal_state_dict, assert_state_dicts_equal


class DummyMultiEnv(ParallelEnv):
    def __init__(self, observation_spaces, action_spaces):
        self.observation_spaces = observation_spaces
        self.action_spaces = action_spaces
        self.agents = ["agent_0", "agent_1", "other_agent_0"]
        self.possible_agents = ["agent_0", "agent_1", "other_agent_0"]
        self.metadata = None
        self.render_mode = None

    def action_space(self, agent):
        return Discrete(self.action_spaces[0].n)

    def observation_space(self, agent):
        return Box(0, 1, self.observation_spaces.shape)

    def reset(self, seed=None, options=None):
        return {
            agent: np.random.rand(*self.observation_spaces.shape)
            for agent in self.agents
        }, {
            "agent_0": {"env_defined_actions": np.array([1])},
            "agent_1": {"env_defined_actions": np.array([1])},
            "other_agent_0": {"env_defined_actions": None},
        }

    def step(self, action):
        return (
            {
                agent: np.random.rand(*self.observation_spaces.shape)
                for agent in self.agents
            },
            {agent: np.random.randint(0, 5) for agent in self.agents},
            {agent: 1 for agent in self.agents},
            {agent: np.random.randint(0, 2) for agent in self.agents},
            self.reset()[1],
        )


class MultiAgentCNNActor(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv3d(
            in_channels=3, out_channels=16, kernel_size=(1, 3, 3), stride=4
        )
        self.conv2 = nn.Conv3d(
            in_channels=16, out_channels=32, kernel_size=(1, 3, 3), stride=2
        )
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.fc1 = nn.Linear(288, 256)
        self.fc2 = nn.Linear(256, 2)
        self.relu = nn.ReLU()
        self.output_activation = GumbelSoftmax()

    def forward(self, state_tensor):
        x = self.relu(self.conv1(state_tensor))
        x = self.relu(self.conv2(x))
        x = x.view(x.size(0), -1)
        x = self.relu(self.fc1(x))
        x = self.output_activation(self.fc2(x))

        return x


class MultiAgentCNNCritic(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv3d(
            in_channels=3, out_channels=16, kernel_size=(2, 3, 3), stride=4
        )
        self.conv2 = nn.Conv3d(
            in_channels=16, out_channels=32, kernel_size=(1, 3, 3), stride=2
        )
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.fc1 = nn.Linear(290, 256)
        self.fc2 = nn.Linear(256, 2)
        self.relu = nn.ReLU()

    def forward(self, state_tensor, action_tensor):
        x = self.relu(self.conv1(state_tensor))
        x = self.relu(self.conv2(x))
        x = x.view(x.size(0), -1)
        x = torch.cat([x, action_tensor], dim=1)
        x = self.relu(self.fc1(x))
        x = self.fc2(x)

        return x


class DummyContinuousQNetwork(ContinuousQNetwork):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, *args, **kwargs):
        return super().forward(*args, **kwargs)

    def no_sync(self):
        class DummyNoSync:
            def __enter__(self):
                return self

            def __exit__(self, exc_type, exc_value, traceback):
                pass  # Add cleanup or handling if needed

        return DummyNoSync()


class DummyDeterministicActor(DeterministicActor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, *args, **kwargs):
        return super().forward(*args, **kwargs)

    def no_sync(self):
        class DummyNoSync:
            def __enter__(self):
                return self

            def __exit__(self, exc_type, exc_value, traceback):
                pass  # Add cleanup or handling if needed

        return DummyNoSync()


@pytest.fixture(scope="function")
def mlp_actor(observation_spaces, action_spaces, request):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    return nn.Sequential(
        nn.Linear(observation_spaces[0].shape[0], 64),
        nn.ReLU(),
        nn.Linear(64, action_spaces[0].n),
        GumbelSoftmax(),
    )


@pytest.fixture(scope="function")
def mlp_critic(action_spaces, observation_spaces, request):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    return nn.Sequential(
        nn.Linear(observation_spaces[0].shape[0] + action_spaces[0].n, 64),
        nn.ReLU(),
        nn.Linear(64, 1),
    )


@pytest.fixture(scope="function")
def cnn_actor():
    return MultiAgentCNNActor()


@pytest.fixture(scope="function")
def cnn_critic():
    return MultiAgentCNNCritic()


@pytest.fixture
def mocked_accelerator():
    MagicMock(spec=Accelerator)


@pytest.fixture(scope="function")
def accelerated_experiences(
    batch_size, observation_spaces, action_spaces, agent_ids, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    one_hot = all(isinstance(space, Discrete) for space in observation_spaces)
    discrete_actions = all(isinstance(space, Discrete) for space in action_spaces)
    state_size = (
        observation_spaces[0].shape if not one_hot else (observation_spaces[0].n,)
    )
    action_size = action_spaces[0].n if discrete_actions else action_spaces[0].shape[0]
    if one_hot:
        states = {
            agent: torch.randint(0, state_size[0], (batch_size, 1)).float()
            for agent in agent_ids
        }
    else:
        states = {agent: torch.randn(batch_size, *state_size) for agent in agent_ids}

    actions = {agent: torch.randn(batch_size, action_size) for agent in agent_ids}
    rewards = {agent: torch.randn(batch_size, 1) for agent in agent_ids}
    dones = {agent: torch.randint(0, 2, (batch_size, 1)) for agent in agent_ids}
    if one_hot:
        next_states = {
            agent: torch.randint(0, state_size[0], (batch_size, 1)).float()
            for agent in agent_ids
        }
    else:
        next_states = {
            agent: torch.randn(batch_size, *state_size) for agent in agent_ids
        }

    return states, actions, rewards, next_states, dones


@pytest.fixture(scope="function")
def experiences(
    batch_size, observation_spaces, action_spaces, agent_ids, device, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    one_hot = all(isinstance(space, Discrete) for space in observation_spaces)
    discrete_actions = all(isinstance(space, Discrete) for space in action_spaces)
    state_size = (
        observation_spaces[0].shape if not one_hot else (observation_spaces[0].n,)
    )
    action_size = action_spaces[0].n if discrete_actions else action_spaces[0].shape[0]
    if one_hot:
        states = {
            agent: torch.randint(0, state_size[0], (batch_size, 1)).float().to(device)
            for agent in agent_ids
        }
    else:
        states = {
            agent: torch.randn(batch_size, *state_size).to(device)
            for agent in agent_ids
        }

    actions = {
        agent: torch.randn(batch_size, action_size).to(device) for agent in agent_ids
    }
    rewards = {agent: torch.randn(batch_size, 1).to(device) for agent in agent_ids}
    dones = {
        agent: torch.randint(0, 2, (batch_size, 1)).to(device) for agent in agent_ids
    }
    if one_hot:
        next_states = {
            agent: torch.randint(0, state_size[0], (batch_size, 1)).float().to(device)
            for agent in agent_ids
        }
    else:
        next_states = {
            agent: torch.randn(batch_size, *state_size).to(device)
            for agent in agent_ids
        }

    return states, actions, rewards, next_states, dones


@pytest.mark.parametrize(
    "observation_spaces",
    [
        "ma_vector_space",
        "ma_image_space",
        "ma_dict_space",
        "ma_multidiscrete_space",
    ],
)
@pytest.mark.parametrize("accelerator_flag", [False, True])
@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_initialize_maddpg_with_net_config(
    accelerator_flag, observation_spaces, ma_vector_space, device, compile_mode, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    net_config = {
        "encoder_config": get_default_encoder_config(observation_spaces[0]),
        "head_config": {"hidden_size": [16]},
    }
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    expl_noise = 0.1
    batch_size = 64
    accelerator = Accelerator() if accelerator_flag else None

    # Initialize MADDPG
    maddpg = MADDPG(
        observation_spaces=observation_spaces,
        net_config=net_config,
        action_spaces=ma_vector_space,
        agent_ids=agent_ids,
        accelerator=accelerator,
        device=device,
        torch_compiler=compile_mode,
    )

    # Check if the MADDPG agent is initialized correctly
    net_config.update({"output_activation": "Softmax"})
    assert maddpg.observation_spaces == observation_spaces
    assert maddpg.action_spaces == ma_vector_space
    assert maddpg.n_agents == len(agent_ids)
    assert maddpg.agent_ids == agent_ids
    for noise_vec in maddpg.expl_noise.values():
        assert torch.all(noise_vec == expl_noise)
    assert maddpg.batch_size == batch_size
    assert maddpg.scores == []
    assert maddpg.fitness == []
    assert maddpg.steps == [0]

    if compile_mode is not None and accelerator is None:
        assert all(
            isinstance(actor, OptimizedModule) for actor in maddpg.actors.values()
        )
        assert all(
            isinstance(critic, OptimizedModule) for critic in maddpg.critics.values()
        )
    else:
        for agent_id in maddpg.agent_ids:
            actor = maddpg.actors[agent_id]
            critic = maddpg.critics[agent_id]
            assert isinstance(actor, DeterministicActor)
            assert isinstance(critic, ContinuousQNetwork)

    expected_optimizer_cls = optim.Adam if accelerator is None else AcceleratedOptimizer
    for agent_id, actor_optimizer in maddpg.actor_optimizers.items():
        critic_optimizer = maddpg.critic_optimizers[agent_id]
        assert isinstance(actor_optimizer, expected_optimizer_cls)
        assert isinstance(critic_optimizer, expected_optimizer_cls)

    assert isinstance(maddpg.criterion, nn.MSELoss)
    maddpg.clean_up()


# TODO: This will be deprecated in the future
@pytest.mark.parametrize("accelerator_flag", [False, True])
@pytest.mark.parametrize("compile_mode", [None, "default"])
@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space"])
def test_initialize_maddpg_with_mlp_networks(
    mlp_actor,
    mlp_critic,
    observation_spaces,
    action_spaces,
    request,
    accelerator_flag,
    device,
    compile_mode,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    accelerator = Accelerator() if accelerator_flag else None
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    evo_actors = ModuleDict(
        {
            agent_id: MakeEvolvable(
                network=mlp_actor, input_tensor=torch.randn(1, 6), device=device
            )
            for agent_id in agent_ids
        }
    )
    evo_critics = ModuleDict(
        {
            agent_id: MakeEvolvable(
                network=mlp_critic, input_tensor=torch.randn(1, 8), device=device
            )
            for agent_id in agent_ids
        }
    )
    maddpg = MADDPG(
        observation_spaces=observation_spaces,
        action_spaces=action_spaces,
        agent_ids=agent_ids,
        actor_networks=evo_actors,
        critic_networks=evo_critics,
        device=device,
        accelerator=accelerator,
        torch_compiler=compile_mode,
    )

    if compile_mode is not None and accelerator is None:
        assert all(
            isinstance(actor, OptimizedModule) for actor in maddpg.actors.values()
        )
        assert all(
            isinstance(critic, OptimizedModule) for critic in maddpg.critics.values()
        )
    else:
        for agent_id in maddpg.agent_ids:
            actor = maddpg.actors[agent_id]
            critic = maddpg.critics[agent_id]
            assert isinstance(actor, MakeEvolvable)
            assert isinstance(critic, MakeEvolvable)

    assert maddpg.observation_spaces == observation_spaces
    assert maddpg.action_spaces == action_spaces
    assert maddpg.n_agents == len(agent_ids)
    assert maddpg.agent_ids == agent_ids
    assert maddpg.scores == []
    assert maddpg.fitness == []
    assert maddpg.steps == [0]
    expected_optimizer_cls = optim.Adam if accelerator is None else AcceleratedOptimizer
    assert all(
        isinstance(actor_optimizer, expected_optimizer_cls)
        for actor_optimizer in maddpg.actor_optimizers.values()
    )
    assert all(
        isinstance(critic_optimizer, expected_optimizer_cls)
        for critic_optimizer in maddpg.critic_optimizers.values()
    )
    assert isinstance(maddpg.criterion, nn.MSELoss)
    maddpg.clean_up()


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space"])
def test_initialize_maddpg_with_mlp_networks_gumbel_softmax(
    mlp_actor,
    mlp_critic,
    observation_spaces,
    action_spaces,
    device,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    compile_mode = "reduce-overhead"
    net_config = {
        "encoder_config": {
            "hidden_size": [64, 64],
            "min_hidden_layers": 1,
            "max_hidden_layers": 3,
            "min_mlp_nodes": 64,
            "max_mlp_nodes": 500,
            "output_activation": "GumbelSoftmax",
            "activation": "ReLU",
        }
    }
    maddpg = MADDPG(
        observation_spaces=observation_spaces,
        action_spaces=action_spaces,
        agent_ids=["agent_0", "agent_1", "other_agent_0"],
        net_config=net_config,
        device=device,
        torch_compiler=compile_mode,
    )
    assert maddpg.torch_compiler == "default"
    maddpg.clean_up()


# TODO: This will be deprecated in the future
@pytest.mark.parametrize("accelerator_flag", [False, True])
@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_initialize_maddpg_with_cnn_networks(
    cnn_actor,
    cnn_critic,
    accelerator_flag,
    device,
    compile_mode,
    ma_image_space,
    ma_discrete_space,
):
    accelerator = Accelerator() if accelerator_flag else None
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    evo_actors = ModuleDict(
        {
            agent_id: MakeEvolvable(
                network=cnn_actor,
                input_tensor=torch.randn(1, 3, 1, 32, 32),
                device=device,
            )
            for agent_id in agent_ids
        }
    )
    evo_critics = ModuleDict(
        {
            agent_id: MakeEvolvable(
                network=cnn_critic,
                input_tensor=torch.randn(1, 3, 3, 32, 32),
                secondary_input_tensor=torch.randn(1, 2),
                device=device,
            )
            for agent_id in agent_ids
        }
    )
    maddpg = MADDPG(
        observation_spaces=ma_image_space,
        action_spaces=ma_discrete_space,
        agent_ids=agent_ids,
        actor_networks=evo_actors,
        critic_networks=evo_critics,
        device=device,
        accelerator=accelerator,
        torch_compiler=compile_mode,
    )
    if compile_mode is not None and accelerator is None:
        assert all(
            isinstance(actor, OptimizedModule) for actor in maddpg.actors.values()
        )
        assert all(
            isinstance(critic, OptimizedModule) for critic in maddpg.critics.values()
        )
    else:
        for agent_id in maddpg.agent_ids:
            actor = maddpg.actors[agent_id]
            critic = maddpg.critics[agent_id]
            assert isinstance(actor, MakeEvolvable)
            assert isinstance(critic, MakeEvolvable)

    assert maddpg.observation_spaces == ma_image_space
    assert maddpg.action_spaces == ma_discrete_space
    assert maddpg.n_agents == len(agent_ids)
    assert maddpg.agent_ids == agent_ids
    assert maddpg.scores == []
    assert maddpg.fitness == []
    assert maddpg.steps == [0]
    expected_optimizer_cls = optim.Adam if accelerator is None else AcceleratedOptimizer
    assert all(
        isinstance(actor_optimizer, expected_optimizer_cls)
        for actor_optimizer in maddpg.actor_optimizers.values()
    )
    assert all(
        isinstance(critic_optimizer, expected_optimizer_cls)
        for critic_optimizer in maddpg.critic_optimizers.values()
    )
    assert isinstance(maddpg.criterion, nn.MSELoss)
    maddpg.clean_up()


@pytest.mark.parametrize("accelerator", [None, Accelerator()])
@pytest.mark.parametrize("compile_mode", [None, "default"])
@pytest.mark.parametrize(
    "observation_spaces, encoder_cls",
    [
        ("ma_vector_space", EvolvableMLP),
        ("ma_image_space", EvolvableCNN),
    ],
)
def test_initialize_maddpg_with_evo_networks(
    observation_spaces,
    encoder_cls,
    device,
    compile_mode,
    accelerator,
    ma_discrete_space,
    request,
):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    observation_spaces = request.getfixturevalue(observation_spaces)
    observation_space = spaces.Dict(
        {agent_id: observation_spaces[idx] for idx, agent_id in enumerate(agent_ids)}
    )

    evo_actors = ModuleDict(
        {
            agent_id: DeterministicActor(
                observation_spaces[i],
                ma_discrete_space[i],
                device=device,
            )
            for i, agent_id in enumerate(agent_ids)
        }
    )
    evo_critics = ModuleDict(
        {
            agent_id: ContinuousQNetwork(
                observation_space=observation_space,
                action_space=concatenate_spaces(ma_discrete_space),
                device=device,
            )
            for agent_id in agent_ids
        }
    )

    maddpg = MADDPG(
        observation_spaces=observation_spaces,
        action_spaces=ma_discrete_space,
        agent_ids=agent_ids,
        actor_networks=evo_actors,
        critic_networks=evo_critics,
        device=device,
        torch_compiler=compile_mode,
        accelerator=accelerator,
    )
    if compile_mode is not None and accelerator is None:
        assert all(
            isinstance(actor, OptimizedModule) for actor in maddpg.actors.values()
        )
        assert all(
            isinstance(critic, OptimizedModule) for critic in maddpg.critics.values()
        )
    else:
        for agent_id in maddpg.agent_ids:
            actor = maddpg.actors[agent_id]
            critic = maddpg.critics[agent_id]
            assert isinstance(actor.encoder, encoder_cls)
            assert isinstance(critic.encoder, EvolvableMultiInput)

    assert maddpg.observation_spaces == observation_spaces
    assert maddpg.action_spaces == ma_discrete_space
    assert maddpg.n_agents == len(agent_ids)
    assert maddpg.agent_ids == agent_ids
    assert maddpg.scores == []
    assert maddpg.fitness == []
    assert maddpg.steps == [0]

    expected_optimizer_cls = optim.Adam if accelerator is None else AcceleratedOptimizer
    assert all(
        isinstance(actor_optimizer, expected_optimizer_cls)
        for actor_optimizer in maddpg.actor_optimizers.values()
    )
    assert all(
        isinstance(critic_optimizer, expected_optimizer_cls)
        for critic_optimizer in maddpg.critic_optimizers.values()
    )
    assert isinstance(maddpg.criterion, nn.MSELoss)
    maddpg.clean_up()


@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_initialize_maddpg_with_incorrect_evo_networks(
    compile_mode, ma_vector_space, ma_discrete_space
):
    evo_actors = []
    evo_critics = []
    with pytest.raises(AssertionError):
        MADDPG(
            observation_spaces=ma_vector_space,
            action_spaces=ma_discrete_space,
            agent_ids=["agent_0", "agent_1", "other_agent_0"],
            actor_networks=evo_actors,
            critic_networks=evo_critics,
            torch_compiler=compile_mode,
        )


@pytest.mark.parametrize("compile_mode", [None, "default"])
@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space"])
def test_maddpg_init_warning(
    mlp_actor, device, compile_mode, observation_spaces, action_spaces, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    warning_string = "Actor and critic network must both be supplied to use custom networks. Defaulting to net config."
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    evo_actors = [
        MakeEvolvable(network=mlp_actor, input_tensor=torch.randn(1, 6), device=device)
        for _ in range(len(agent_ids))
    ]
    with pytest.warns(UserWarning, match=warning_string):
        MADDPG(
            observation_spaces=observation_spaces,
            action_spaces=action_spaces,
            agent_ids=agent_ids,
            actor_networks=evo_actors,
            device=device,
            torch_compiler=compile_mode,
        )


@pytest.mark.parametrize(
    "mode", (None, 0, False, "default", "reduce-overhead", "max-autotune")
)
def test_maddpg_init_torch_compiler_no_error(
    mode, ma_vector_space, ma_discrete_space, device
):
    maddpg = MADDPG(
        observation_spaces=ma_vector_space,
        action_spaces=ma_discrete_space,
        agent_ids=["agent_0", "agent_1", "other_agent_0"],
        device=device,
        torch_compiler=mode,
    )
    if isinstance(mode, str):
        assert all(
            isinstance(actor, OptimizedModule) for actor in maddpg.actors.values()
        )
        assert all(
            isinstance(critic, OptimizedModule) for critic in maddpg.critics.values()
        )
        assert all(
            isinstance(actor_target, OptimizedModule)
            for actor_target in maddpg.actor_targets.values()
        )
        assert all(
            isinstance(critic_target, OptimizedModule)
            for critic_target in maddpg.critic_targets.values()
        )
        assert maddpg.torch_compiler == "default"
    else:
        assert isinstance(maddpg, MADDPG)
    maddpg.clean_up()


@pytest.mark.parametrize("mode", (1, True, "max-autotune-no-cudagraphs"))
def test_maddpg_init_torch_compiler_error(
    mode, ma_vector_space, ma_discrete_space, device
):
    err_string = (
        "Choose between torch compiler modes: "
        "default, reduce-overhead, max-autotune or None"
    )
    with pytest.raises(AssertionError, match=err_string):
        MADDPG(
            observation_spaces=ma_vector_space,
            action_spaces=ma_discrete_space,
            agent_ids=["agent_0", "agent_1", "other_agent_0"],
            device=device,
            torch_compiler=mode,
        )


@pytest.mark.parametrize(
    "observation_spaces", ["ma_vector_space", "ma_discrete_space", "ma_image_space"]
)
@pytest.mark.parametrize("action_spaces", ["ma_vector_space", "ma_discrete_space"])
@pytest.mark.parametrize("training", [0, 1])
@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_maddpg_get_action(
    training, observation_spaces, action_spaces, device, compile_mode, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    if all(isinstance(space, spaces.Discrete) for space in observation_spaces):
        state = {
            agent: np.random.randint(0, observation_spaces[idx].n, 1)
            for idx, agent in enumerate(agent_ids)
        }
    else:
        state = {
            agent: np.random.randn(*observation_spaces[idx].shape)
            for idx, agent in enumerate(agent_ids)
        }

    maddpg = MADDPG(
        observation_spaces,
        action_spaces,
        agent_ids=agent_ids,
        device=device,
        torch_compiler=compile_mode,
    )
    maddpg.set_training_mode(bool(training))
    processed_action, raw_action = maddpg.get_action(state)
    discrete_actions = all(
        isinstance(space, spaces.Discrete) for space in action_spaces
    )
    for idx, env_actions in enumerate(list(raw_action.values())):
        action_dim = (
            action_spaces[idx].shape[0]
            if isinstance(action_spaces[idx], spaces.Box)
            else action_spaces[idx].n
        )
        for action in env_actions:
            assert len(action) == action_dim
            if discrete_actions:
                torch.testing.assert_close(
                    sum(action),
                    1.0,
                    atol=0.1,
                    rtol=1e-3,
                )
            assert action.dtype == np.float32
            assert -1 <= action.all() <= 1

    if discrete_actions:
        for idx, env_action in enumerate(list(processed_action.values())):
            for action in env_action:
                assert action <= action_spaces[idx].n - 1
    maddpg.clean_up()
    maddpg = None


@pytest.mark.parametrize("training", [False, True])
def test_maddpg_get_action_action_masking_exception(
    training, device, ma_vector_space, ma_discrete_space
):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    state = {
        agent: {
            "observation": np.random.randn(*ma_vector_space[idx].shape),
            "action_mask": [0, 1, 0, 1],
        }
        for idx, agent in enumerate(agent_ids)
    }
    maddpg = MADDPG(
        ma_vector_space,
        ma_discrete_space,
        agent_ids=agent_ids,
        device=device,
    )
    with pytest.raises(AssertionError):
        maddpg.set_training_mode(training)
        _, raw_action = maddpg.get_action(state)
    maddpg.clean_up()


@pytest.mark.parametrize("training", [False, True])
def test_maddpg_get_action_action_masking(
    training, device, ma_vector_space, ma_discrete_space
):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    state = {
        agent: np.random.randn(*ma_vector_space[idx].shape)
        for idx, agent in enumerate(agent_ids)
    }
    info = {
        agent: {
            "action_mask": [0, 1],
        }
        for idx, agent in enumerate(agent_ids)
    }
    maddpg = MADDPG(
        ma_vector_space,
        ma_discrete_space,
        agent_ids=agent_ids,
        device=device,
    )
    maddpg.set_training_mode(training)
    action, _ = maddpg.get_action(state, info)
    assert all(i in [1, 3] for i in action.values())
    maddpg.clean_up()


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space", "ma_image_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space", "ma_vector_space"])
@pytest.mark.parametrize("training", [False, True])
@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_get_action_distributed(
    training, observation_spaces, action_spaces, compile_mode, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    accelerator = Accelerator()
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    state = {
        agent: np.random.randn(*observation_spaces[idx].shape)
        for idx, agent in enumerate(agent_ids)
    }
    from agilerl.algorithms.maddpg import MADDPG

    maddpg = MADDPG(
        observation_spaces,
        action_spaces,
        agent_ids=agent_ids,
        accelerator=accelerator,
        torch_compiler=compile_mode,
    )
    new_actors = ModuleDict(
        {
            agent_id: DummyDeterministicActor(
                observation_space=actor.observation_space,
                action_space=actor.action_space,
                encoder_config=actor.encoder.net_config,
                head_config=actor.head_net.net_config,
                device=actor.device,
            )
            for agent_id, actor in maddpg.actors.items()
        }
    )
    maddpg.actors = new_actors
    maddpg.set_training_mode(training)
    processed_action, raw_action = maddpg.get_action(state)
    discrete_actions = all(
        isinstance(space, spaces.Discrete) for space in action_spaces
    )
    for idx, env_actions in enumerate(list(raw_action.values())):
        action_dim = (
            action_spaces[idx].shape[0]
            if isinstance(action_spaces[idx], spaces.Box)
            else action_spaces[idx].n
        )
        for action in env_actions:
            assert len(action) == action_dim
            if discrete_actions:
                torch.testing.assert_close(
                    sum(action),
                    1.0,
                    atol=0.1,
                    rtol=1e-3,
                )
            assert action.dtype == np.float32
            assert -1 <= action.all() <= 1

    if discrete_actions:
        for idx, env_action in enumerate(list(processed_action.values())):
            action_dim = (
                action_spaces[idx].shape[0]
                if isinstance(action_spaces[idx], spaces.Box)
                else action_spaces[idx].n
            )
            for action in env_action:
                assert action <= action_dim - 1
    maddpg.clean_up()


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_vector_space", "ma_discrete_space"])
@pytest.mark.parametrize("training", [False, True])
@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_maddpg_get_action_agent_masking(
    training, observation_spaces, action_spaces, device, compile_mode, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    state = {
        agent: np.random.randn(*observation_spaces[0].shape) for agent in agent_ids
    }
    discrete_actions = all(
        isinstance(space, spaces.Discrete) for space in action_spaces
    )
    if discrete_actions:
        info = {
            "agent_0": {"env_defined_actions": 1},
            "agent_1": {"env_defined_actions": 1},
            "other_agent_0": {"env_defined_actions": None},
        }
    else:
        info = {
            "agent_0": {"env_defined_actions": np.array([0, 1, 0, 1, 0, 1])},
            "agent_1": {"env_defined_actions": np.array([0, 1, 0, 1, 0, 1])},
            "other_agent_0": {"env_defined_actions": None},
        }
    maddpg = MADDPG(
        observation_spaces,
        action_spaces,
        agent_ids=agent_ids,
        device=device,
        torch_compiler=compile_mode,
    )
    maddpg.set_training_mode(training)
    action, _ = maddpg.get_action(state, infos=info)
    if discrete_actions:
        assert np.array_equal(action["agent_0"], np.array([1])), action["agent_0"]
    else:
        assert np.array_equal(
            action["agent_0"], np.array([[0, 1, 0, 1, 0, 1]])
        ), action["agent_0"]
    maddpg.clean_up()


@pytest.mark.parametrize("training", [False, True])
@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_vector_space", "ma_discrete_space"])
def test_maddpg_get_action_vectorized_agent_masking(
    training, observation_spaces, action_spaces, device, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    num_envs = 6
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    state = {
        agent: np.array(
            [np.random.randn(*observation_spaces[0].shape) for _ in range(num_envs)]
        )
        for agent in agent_ids
    }
    discrete_actions = all(
        isinstance(space, spaces.Discrete) for space in action_spaces
    )
    if discrete_actions:
        env_defined_action = np.array(
            [
                np.random.randint(0, observation_spaces[0].shape[0] + 1)
                for _ in range(num_envs)
            ]
        )
    else:
        env_defined_action = np.array(
            [np.random.randn(*observation_spaces[0].shape) for _ in range(num_envs)]
        )
    nan_array = np.zeros(env_defined_action.shape)
    nan_array[:] = np.nan
    info = {
        "agent_0": {"env_defined_actions": env_defined_action},
        "agent_1": {"env_defined_actions": env_defined_action},
        "other_agent_0": {"env_defined_actions": nan_array},
    }
    maddpg = MADDPG(
        observation_spaces,
        action_spaces,
        agent_ids=agent_ids,
        device=device,
    )
    maddpg.set_training_mode(training)
    action, raw_action = maddpg.get_action(state, infos=info)
    if discrete_actions:
        assert np.array_equal(
            action["agent_0"].squeeze(), info["agent_0"]["env_defined_actions"]
        ), action["agent_0"]
    else:
        assert np.isclose(
            action["agent_0"], info["agent_0"]["env_defined_actions"]
        ).all(), action["agent_0"]
    maddpg.clean_up()


@pytest.mark.parametrize(
    "observation_spaces", ["ma_vector_space", "ma_discrete_space", "ma_image_space"]
)
@pytest.mark.parametrize("action_spaces", ["ma_vector_space", "ma_discrete_space"])
@pytest.mark.parametrize("agent_ids", [["agent_0", "agent_1", "other_agent_0"]])
@pytest.mark.parametrize("compile_mode", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("batch_size", [64])
def test_maddpg_learns_from_experiences(
    observation_spaces,
    action_spaces,
    experiences,
    batch_size,
    agent_ids,
    device,
    compile_mode,
    accelerator,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    maddpg = MADDPG(
        observation_spaces,
        action_spaces,
        agent_ids=agent_ids,
        device=device,
        accelerator=accelerator,
        torch_compiler=compile_mode,
    )
    if accelerator is not None:
        for agent_id, actor in maddpg.actors.items():
            critic = maddpg.critics[agent_id]
            actor_target = maddpg.actor_targets[agent_id]
            critic_target = maddpg.critic_targets[agent_id]
            actor.no_sync = no_sync.__get__(actor)
            critic.no_sync = no_sync.__get__(critic)
            actor_target.no_sync = no_sync.__get__(actor_target)
            critic_target.no_sync = no_sync.__get__(critic_target)

    actors_pre_learn_sd = {
        agent_id: copy.deepcopy(actor.state_dict())
        for agent_id, actor in maddpg.actors.items()
    }
    critics_pre_learn_sd = {
        agent_id: copy.deepcopy(critic.state_dict())
        for agent_id, critic in maddpg.critics.items()
    }

    for _ in range(2):
        maddpg.scores.append(0)
        loss = maddpg.learn(experiences)

    assert isinstance(loss, dict)

    for agent_id in maddpg.agent_ids:
        assert loss[agent_id][-1] >= 0.0

        updated_actor = maddpg.actors[agent_id]
        assert_not_equal_state_dict(
            actors_pre_learn_sd[agent_id], updated_actor.state_dict()
        )

        updated_critic = maddpg.critics[agent_id]
        assert_not_equal_state_dict(
            critics_pre_learn_sd[agent_id], updated_critic.state_dict()
        )
    maddpg.clean_up()


def no_sync(self):
    class DummyNoSync:
        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            pass  # Add cleanup or handling if needed

    return DummyNoSync()


@pytest.mark.parametrize("compile_mode", [None])
def test_maddpg_soft_update(device, compile_mode, ma_vector_space, ma_discrete_space):
    maddpg = MADDPG(
        ma_vector_space,
        ma_discrete_space,
        agent_ids=["agent_0", "agent_1", "other_agent_0"],
        accelerator=None,
        device=device,
        torch_compiler=compile_mode,
    )

    for agent_id, actor in maddpg.actors.items():
        actor_target = maddpg.actor_targets[agent_id]
        # Check actors
        maddpg.soft_update(actor, actor_target)
        eval_params = list(actor.parameters())
        target_params = list(actor_target.parameters())
        expected_params = [
            maddpg.tau * eval_param + (1.0 - maddpg.tau) * target_param
            for eval_param, target_param in zip(eval_params, target_params)
        ]
        assert all(
            torch.allclose(expected_param, target_param)
            for expected_param, target_param in zip(expected_params, target_params)
        )

    for agent_id, critic in maddpg.critics.items():
        critic_target = maddpg.critic_targets[agent_id]
        maddpg.soft_update(critic, critic_target)
        eval_params = list(critic.parameters())
        target_params = list(critic_target.parameters())
        expected_params = [
            maddpg.tau * eval_param + (1.0 - maddpg.tau) * target_param
            for eval_param, target_param in zip(eval_params, target_params)
        ]

        assert all(
            torch.allclose(expected_param, target_param)
            for expected_param, target_param in zip(expected_params, target_params)
        )
    maddpg.clean_up()


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space", "ma_image_space"])
@pytest.mark.parametrize("sum_score", [True, False])
@pytest.mark.parametrize("compile_mode", [None])
@pytest.mark.parametrize("vectorized", [True, False])
def test_maddpg_algorithm_test_loop(
    ma_discrete_space,
    device,
    sum_score,
    compile_mode,
    observation_spaces,
    vectorized,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)

    # Define environment and algorithm
    if vectorized:
        env = make_multi_agent_vect_envs(
            DummyMultiEnv,
            2,
            **dict(
                observation_spaces=observation_spaces[0],
                action_spaces=ma_discrete_space,
            ),
        )
    else:
        env = DummyMultiEnv(observation_spaces[0], ma_discrete_space)

    maddpg = MADDPG(
        observation_spaces,
        ma_discrete_space,
        agent_ids=["agent_0", "agent_1", "other_agent_0"],
        device=device,
        torch_compiler=compile_mode,
    )
    mean_score = maddpg.test(env, max_steps=10, sum_scores=sum_score)
    if sum_score:
        assert isinstance(mean_score, float)
    else:
        assert isinstance(mean_score, np.ndarray)
        assert len(mean_score) == 3
    env.close()
    maddpg.clean_up()


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("compile_mode", [None, "default"])
@pytest.mark.parametrize("accelerator_flag", [False, True])
@pytest.mark.parametrize("wrap", [True, False])
def test_maddpg_clone_returns_identical_agent(
    accelerator_flag, wrap, compile_mode, observation_spaces, ma_vector_space, request
):
    # Clones the agent and returns an identical copy.
    observation_spaces = request.getfixturevalue(observation_spaces)
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    expl_noise = 0.1
    index = 0
    batch_size = 64
    lr_actor = 0.001
    lr_critic = 0.01
    learn_step = 5
    gamma = 0.95
    tau = 0.01
    mut = None
    actor_networks = None
    critic_networks = None
    device = "cpu"
    accelerator = Accelerator(device_placement=False) if accelerator_flag else None

    maddpg = MADDPG(
        observation_spaces,
        ma_vector_space,
        agent_ids,
        expl_noise=expl_noise,
        index=index,
        batch_size=batch_size,
        lr_actor=lr_actor,
        lr_critic=lr_critic,
        learn_step=learn_step,
        gamma=gamma,
        tau=tau,
        mut=mut,
        actor_networks=actor_networks,
        critic_networks=critic_networks,
        device=device,
        accelerator=accelerator,
        wrap=wrap,
        torch_compiler=compile_mode,
    )
    clone_agent = maddpg.clone(wrap=wrap)

    assert isinstance(clone_agent, MADDPG)
    assert clone_agent.observation_spaces == maddpg.observation_spaces
    assert clone_agent.action_spaces == maddpg.action_spaces
    assert clone_agent.n_agents == maddpg.n_agents
    assert clone_agent.agent_ids == maddpg.agent_ids
    assert all(
        torch.equal(clone_expl_noise, expl_noise)
        for clone_expl_noise, expl_noise in zip(
            clone_agent.expl_noise.values(), maddpg.expl_noise.values()
        )
    )
    assert clone_agent.index == maddpg.index
    assert clone_agent.batch_size == maddpg.batch_size
    assert clone_agent.lr_actor == maddpg.lr_actor
    assert clone_agent.lr_critic == maddpg.lr_critic
    assert clone_agent.learn_step == maddpg.learn_step
    assert clone_agent.gamma == maddpg.gamma
    assert clone_agent.tau == maddpg.tau
    assert clone_agent.device == maddpg.device
    assert clone_agent.accelerator == maddpg.accelerator

    for agent_id in clone_agent.agent_ids:
        actor = maddpg.actors[agent_id]
        clone_actor = clone_agent.actors[agent_id]
        assert_state_dicts_equal(clone_actor.state_dict(), actor.state_dict())

        critic = maddpg.critics[agent_id]
        clone_critic = clone_agent.critics[agent_id]
        assert_state_dicts_equal(clone_critic.state_dict(), critic.state_dict())

        actor_target = maddpg.actor_targets[agent_id]
        clone_actor_target = clone_agent.actor_targets[agent_id]
        assert_state_dicts_equal(
            clone_actor_target.state_dict(), actor_target.state_dict()
        )

        critic_target = maddpg.critic_targets[agent_id]
        clone_critic_target = clone_agent.critic_targets[agent_id]
        assert_state_dicts_equal(
            clone_critic_target.state_dict(), critic_target.state_dict()
        )
    maddpg.clean_up()
    clone_agent.clean_up()


@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_clone_new_index(compile_mode, ma_vector_space, ma_discrete_space):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]

    maddpg = MADDPG(
        ma_vector_space,
        ma_discrete_space,
        agent_ids,
        torch_compiler=compile_mode,
    )
    clone_agent = maddpg.clone(index=100)

    assert clone_agent.index == 100
    maddpg.clean_up()
    clone_agent.clean_up()


@pytest.mark.parametrize("compile_mode", [None])
def test_clone_after_learning(compile_mode, ma_vector_space):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    batch_size = 8

    maddpg = MADDPG(
        ma_vector_space,
        copy.deepcopy(ma_vector_space),
        agent_ids,
        batch_size=batch_size,
        torch_compiler=compile_mode,
    )

    states = {
        agent_id: torch.randn(batch_size, ma_vector_space[idx].shape[0])
        for idx, agent_id in enumerate(agent_ids)
    }
    actions = {
        agent_id: torch.randn(batch_size, ma_vector_space[idx].shape[0])
        for idx, agent_id in enumerate(agent_ids)
    }
    rewards = {agent_id: torch.randn(batch_size, 1) for agent_id in agent_ids}
    next_states = {
        agent_id: torch.randn(batch_size, ma_vector_space[idx].shape[0])
        for idx, agent_id in enumerate(agent_ids)
    }
    dones = {agent_id: torch.zeros(batch_size, 1) for agent_id in agent_ids}

    experiences = states, actions, rewards, next_states, dones
    maddpg.learn(experiences)
    clone_agent = maddpg.clone()
    assert isinstance(clone_agent, MADDPG)
    assert clone_agent.observation_spaces == maddpg.observation_spaces
    assert clone_agent.action_spaces == maddpg.action_spaces
    assert clone_agent.n_agents == maddpg.n_agents
    assert clone_agent.agent_ids == maddpg.agent_ids
    assert all(
        torch.equal(clone_expl_noise, expl_noise)
        for clone_expl_noise, expl_noise in zip(
            clone_agent.expl_noise.values(), maddpg.expl_noise.values()
        )
    )
    assert clone_agent.index == maddpg.index
    assert clone_agent.batch_size == maddpg.batch_size
    assert clone_agent.lr_actor == maddpg.lr_actor
    assert clone_agent.lr_critic == maddpg.lr_critic
    assert clone_agent.learn_step == maddpg.learn_step
    assert clone_agent.gamma == maddpg.gamma
    assert clone_agent.tau == maddpg.tau
    assert clone_agent.device == maddpg.device
    assert clone_agent.accelerator == maddpg.accelerator

    for agent_id in clone_agent.agent_ids:
        actor = maddpg.actors[agent_id]
        clone_actor = clone_agent.actors[agent_id]
        assert_state_dicts_equal(clone_actor.state_dict(), actor.state_dict())

        critic = maddpg.critics[agent_id]
        clone_critic = clone_agent.critics[agent_id]
        assert_state_dicts_equal(clone_critic.state_dict(), critic.state_dict())

        clone_actor_target = clone_agent.actor_targets[agent_id]
        actor_target = maddpg.actor_targets[agent_id]
        assert_state_dicts_equal(
            clone_actor_target.state_dict(), actor_target.state_dict()
        )

        clone_critic_target = clone_agent.critic_targets[agent_id]
        critic_target = maddpg.critic_targets[agent_id]
        assert_state_dicts_equal(
            clone_critic_target.state_dict(), critic_target.state_dict()
        )

        clone_actor_opt = clone_agent.actor_optimizers.optimizer[agent_id]
        actor_opt = maddpg.actor_optimizers.optimizer[agent_id]
        assert str(clone_actor_opt) == str(actor_opt)

        clone_critic_opt = clone_agent.critic_optimizers.optimizer[agent_id]
        critic_opt = maddpg.critic_optimizers.optimizer[agent_id]
        assert str(clone_critic_opt) == str(critic_opt)
    maddpg.clean_up()
    clone_agent.clean_up()



import time
import warnings
from datetime import datetime
from typing import Any, Callable, Optional

import gymnasium as gym
import numpy as np
import wandb
from accelerate import Accelerator
from gymnasium import spaces

from agilerl.algorithms import PPO
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.networks import StochasticActor
from agilerl.utils.algo_utils import obs_channels_to_first
from agilerl.utils.utils import (
    default_progress_bar,
    init_wandb,
    save_population_checkpoint,
    tournament_selection_and_mutation,
)

InitDictType = Optional[dict[str, Any]]
OnPolicyAlgorithms = PPO
PopulationType = list[OnPolicyAlgorithms]


def train_on_policy(
    env: gym.Env,
    env_name: str,
    algo: str,
    pop: PopulationType,
    INIT_HP: InitDictType = None,
    MUT_P: InitDictType = None,
    swap_channels: bool = False,
    max_steps: int = 1000000,
    evo_steps: int = 10000,
    eval_steps: Optional[int] = None,
    eval_loop: int = 1,
    target: Optional[float] = None,
    tournament: Optional[TournamentSelection] = None,
    mutation: Optional[Mutations] = None,
    checkpoint: Optional[int] = None,
    checkpoint_path: Optional[str] = None,
    overwrite_checkpoints: bool = False,
    save_elite: bool = False,
    elite_path: Optional[str] = None,
    wb: bool = False,
    verbose: bool = True,
    accelerator: Optional[Accelerator] = None,
    wandb_api_key: Optional[str] = None,
    wandb_kwargs: Optional[dict[str, Any]] = None,
    collect_rollouts_fn: Optional[
        Callable[[OnPolicyAlgorithms, gym.Env, int], None]
    ] = None,
) -> tuple[PopulationType, list[list[float]]]:
    """The general on-policy RL training function. Returns trained population of agents
    and their fitnesses.

    :param env: The environment to train in. Can be vectorized.
    :type env: Gym-style environment
    :param env_name: Environment name
    :type env_name: str
    :param algo: RL algorithm name
    :type algo: str
    :param pop: Population of agents
    :type pop: list[RLAlgorithm]
    :param INIT_HP: Dictionary containing initial hyperparameters, defaults to None
    :type INIT_HP: dict, optional
    :param MUT_P: Dictionary containing mutation parameters, defaults to None
    :type MUT_P: dict, optional
    :param swap_channels: Swap image channels dimension from last to first
        [H, W, C] -> [C, H, W], defaults to False
    :type swap_channels: bool, optional
    :param max_steps: Maximum number of steps in environment, defaults to 1000000
    :type max_steps: int, optional
    :param evo_steps: Evolution frequency (steps), defaults to 10000
    :type evo_steps: int, optional
    :param eval_steps: Number of evaluation steps per episode. If None, will evaluate until
        environment terminates or truncates. Defaults to None
    :type eval_steps: int, optional
    :param eval_loop: Number of evaluation episodes, defaults to 1
    :type eval_loop: int, optional
    :param target: Target score for early stopping, defaults to None
    :type target: float, optional
    :param tournament: Tournament selection object, defaults to None
    :type tournament: object, optional
    :param mutation: Mutation object, defaults to None
    :type mutation: object, optional
    :param checkpoint: Checkpoint frequency (steps), defaults to None
    :type checkpoint: int, optional
    :param checkpoint_path: Location to save checkpoint, defaults to None
    :type checkpoint_path: str, optional
    :param overwrite_checkpoints: Overwrite previous checkpoints during training, defaults to False
    :type overwrite_checkpoints: bool, optional
    :param save_elite: Boolean flag indicating whether to save elite member at the end
        of training, defaults to False
    :type save_elite: bool, optional
    :param elite_path: Location to save elite agent, defaults to None
    :type elite_path: str, optional
    :param wb: Weights & Biases tracking, defaults to False
    :type wb: bool, optional
    :param verbose: Display training stats, defaults to True
    :type verbose: bool, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    :param wandb_api_key: API key for Weights & Biases, defaults to None
    :type wandb_api_key: str, optional
    :param wandb_kwargs: Additional kwargs to pass to wandb.init()
    :type wand_kwargs: dict, optional
    :param collect_rollouts_fn: Optional function used to collect rollouts. If
        ``None`` and agents use a rollout buffer, a default function will be
        selected based on whether the agent is recurrent.
    :type collect_rollouts_fn: Callable or None, optional

    :return: Trained population of agents and their fitnesses
    :rtype: list[RLAlgorithm], list[list[float]]
    """
    assert isinstance(
        algo, str
    ), "'algo' must be the name of the algorithm as a string."
    assert isinstance(max_steps, int), "Number of steps must be an integer."
    assert isinstance(evo_steps, int), "Evolution frequency must be an integer."
    if target is not None:
        assert isinstance(
            target, (float, int)
        ), "Target score must be a float or an integer."
    if checkpoint is not None:
        assert isinstance(checkpoint, int), "Checkpoint must be an integer."
    assert isinstance(
        wb, bool
    ), "'wb' must be a boolean flag, indicating whether to record run with W&B"
    assert isinstance(verbose, bool), "Verbose must be a boolean."
    if save_elite is False and elite_path is not None:
        warnings.warn(
            "'save_elite' set to False but 'elite_path' has been defined, elite will not\
                      be saved unless 'save_elite' is set to True."
        )
    if checkpoint is None and checkpoint_path is not None:
        warnings.warn(
            "'checkpoint' set to None but 'checkpoint_path' has been defined, checkpoint will not\
                      be saved unless 'checkpoint' is defined."
        )

    if wb:
        init_wandb_kwargs = {
            "algo": algo,
            "env_name": env_name,
            "init_hyperparams": INIT_HP,
            "mutation_hyperparams": MUT_P,
            "wandb_api_key": wandb_api_key,
            "accelerator": accelerator,
        }
        if wandb_kwargs is not None:
            init_wandb_kwargs.update(wandb_kwargs)

        init_wandb(**init_wandb_kwargs)

    # Detect if environment is vectorised
    if hasattr(env, "num_envs"):
        is_vectorised = True
        num_envs = env.num_envs
    else:
        is_vectorised = False
        num_envs = 1

    save_path = (
        checkpoint_path.split(".pt")[0]
        if checkpoint_path is not None
        else "{}-EvoHPO-{}-{}".format(
            env_name, algo, datetime.now().strftime("%m%d%Y%H%M%S")
        )
    )

    if accelerator is not None:
        print(f"\nDistributed training on {accelerator.device}...")
    else:
        print("\nTraining...")

    # Format progress bar
    pbar = default_progress_bar(max_steps, accelerator)

    pop_loss = [[] for _ in pop]
    pop_entropy = [[] for _ in pop]
    pop_fitnesses = []
    total_steps = 0
    loss = None
    checkpoint_count = 0

    # Pre-training mutation
    if accelerator is None and mutation is not None:
        pop = mutation.mutation(pop, pre_training_mut=True)

    # RL training loop
    active_collect = collect_rollouts_fn
    while np.less([agent.steps[-1] for agent in pop], max_steps).all():
        if accelerator is not None:
            accelerator.wait_for_everyone()
        pop_episode_scores = []
        pop_fps = []
        for agent_idx, agent in enumerate(pop):  # Loop through population
            agent.set_training_mode(True)

            completed_episode_scores = []
            steps = 0
            start_time = time.time()

            n_steps = -(agent.learn_step // -num_envs)
            if active_collect is None and getattr(agent, "use_rollout_buffer", False):
                if getattr(agent, "recurrent", False):
                    from agilerl.rollouts import (
                        collect_rollouts_recurrent as active_collect,
                    )
                else:
                    from agilerl.rollouts import collect_rollouts as active_collect

            if (
                getattr(agent, "use_rollout_buffer", False)
                and active_collect is not None
            ):
                last_obs, last_done, last_scores, last_info = None, None, None, None
                for _ in range(-(evo_steps // -agent.learn_step)):
                    # Collect rollouts and save in buffer
                    episode_scores, last_obs, last_done, last_scores, last_info = (
                        active_collect(
                            agent,
                            env,
                            n_steps=n_steps,
                            last_obs=last_obs,
                            last_done=last_done,
                            last_scores=last_scores,
                            last_info=last_info,
                        )
                    )

                    loss = agent.learn()  # Learn from rollout buffer
                    pop_loss[agent_idx].append(loss)

                    # Update step counter and scores
                    steps += n_steps * num_envs
                    total_steps += n_steps * num_envs
                    completed_episode_scores += episode_scores

            # Collect rollouts explicitly without saving to rollout buffer
            else:
                obs, info = env.reset()
                scores = np.zeros(num_envs)
                for _ in range(-(evo_steps // -agent.learn_step)):

                    observations = []
                    actions = []
                    log_probs = []
                    rewards = []
                    dones = []
                    values = []

                    done = np.zeros(num_envs)
                    for _ in range(-(agent.learn_step // -num_envs)):
                        if swap_channels:
                            obs = obs_channels_to_first(obs)

                        action_mask = info.get("action_mask", None)
                        action, log_prob, entropy, value = agent.get_action(
                            obs, action_mask=action_mask
                        )

                        if not is_vectorised:
                            action = action[0]
                            log_prob = log_prob[0]
                            value = value[0]
                            entropy = (
                                entropy[0] if hasattr(entropy, "__len__") else entropy
                            )

                        pop_entropy[agent_idx].append(entropy)

                        # Clip action to action space
                        policy = getattr(agent, agent.registry.policy())
                        if isinstance(policy, StochasticActor) and isinstance(
                            agent.action_space, spaces.Box
                        ):
                            if policy.squash_output:
                                clipped_action = policy.scale_action(action)
                            else:
                                clipped_action = np.clip(
                                    action,
                                    agent.action_space.low,
                                    agent.action_space.high,
                                )
                        else:
                            clipped_action = action

                        next_obs, reward, term, trunc, info = env.step(clipped_action)

                        # Check if termination condition is met
                        if isinstance(term, (list, np.ndarray)):
                            next_done = (
                                np.logical_or(term, trunc).astype(np.int8)
                                if isinstance(trunc, (list, np.ndarray))
                                else term
                            )
                        else:
                            next_done = term or trunc

                        reward_np = np.atleast_1d(reward)
                        next_done_np = np.atleast_1d(next_done)
                        value_np = np.atleast_1d(value)
                        log_prob_np = np.atleast_1d(log_prob)

                        observations.append(obs)
                        actions.append(action)
                        log_probs.append(log_prob_np)
                        rewards.append(reward_np)
                        dones.append(done)
                        values.append(value_np)

                        obs = next_obs
                        done = next_done

                        scores += reward_np
                        total_steps += num_envs
                        steps += num_envs

                        for idx, env_done in enumerate(next_done_np):
                            if env_done:
                                completed_episode_scores.append(scores[idx])
                                agent.scores.append(scores[idx])
                                scores[idx] = 0

                    if swap_channels:
                        next_obs = obs_channels_to_first(next_obs)

                    experiences = (
                        observations,
                        actions,
                        log_probs,
                        rewards,
                        dones,
                        values,
                        next_obs,
                        next_done,
                    )
                    loss = agent.learn(experiences)
                    pop_loss[agent_idx].append(loss)

            agent.steps[-1] += steps
            fps = steps / (time.time() - start_time)
            pop_fps.append(fps)
            pbar.update(steps // len(pop))
            pop_episode_scores.append(completed_episode_scores)

        # Evaluate population
        fitnesses = [
            agent.test(
                env,
                swap_channels=swap_channels,
                max_steps=eval_steps,
                loop=eval_loop,
            )
            for agent in pop
        ]
        pop_fitnesses.append(fitnesses)
        mean_scores = [
            (
                np.mean(episode_scores)
                if len(episode_scores) > 0
                else "0 completed episodes"
            )
            for episode_scores in pop_episode_scores
        ]

        if wb:
            wandb_dict = {
                "global_step": (
                    total_steps * accelerator.state.num_processes
                    if accelerator is not None and accelerator.is_main_process
                    else total_steps
                ),
                "fps": np.mean(pop_fps),
                "train/mean_score": np.mean(
                    [
                        mean_score
                        for mean_score in mean_scores
                        if not isinstance(mean_score, str)
                    ]
                ),
                "eval/mean_fitness": np.mean(fitnesses),
                "eval/best_fitness": np.max(fitnesses),
            }

            agent_loss_dict = {
                f"train/agent_{index}_loss": np.mean(loss_[-10:])
                for index, loss_ in enumerate(pop_loss)
            }
            wandb_dict.update(agent_loss_dict)

            # Add entropy metrics to wandb
            agent_entropy_dict = {
                f"train/agent_{index}_entropy": (
                    np.mean(entropy_[-100:]) if len(entropy_) > 0 else 0
                )
                for index, entropy_ in enumerate(pop_entropy)
            }
            wandb_dict.update(agent_entropy_dict)

            # Add mean entropy across all agents
            all_entropies = [
                np.mean(entropy_[-100:]) if len(entropy_) > 0 else 0
                for entropy_ in pop_entropy
            ]
            if all_entropies:
                wandb_dict["train/mean_entropy"] = np.mean(all_entropies)

            if accelerator is not None:
                accelerator.wait_for_everyone()
                if accelerator.is_main_process:
                    wandb.log(wandb_dict)
                accelerator.wait_for_everyone()
            else:
                wandb.log(wandb_dict)

        # Update step counter
        for agent in pop:
            agent.steps.append(agent.steps[-1])

        # Early stop if consistently reaches target
        if target is not None:
            if (
                np.all(
                    np.greater([np.mean(agent.fitness[-10:]) for agent in pop], target)
                )
                and len(pop[0].steps) >= 100
            ):
                if wb:
                    wandb.finish()
                return pop, pop_fitnesses

        # Tournament selection and population mutation
        if tournament and mutation is not None:
            pop = tournament_selection_and_mutation(
                population=pop,
                tournament=tournament,
                mutation=mutation,
                env_name=env_name,
                algo=algo,
                elite_path=elite_path,
                save_elite=save_elite,
                accelerator=accelerator,
            )

        if verbose:
            fitness = ["%.2f" % fitness for fitness in fitnesses]
            avg_fitness = ["%.2f" % np.mean(agent.fitness[-5:]) for agent in pop]
            avg_score = ["%.2f" % np.mean(agent.scores[-10:]) for agent in pop]
            mean_scores = [
                "%.2f" % mean_score if not isinstance(mean_score, str) else mean_score
                for mean_score in mean_scores
            ]
            agents = [agent.index for agent in pop]
            num_steps = [agent.steps[-1] for agent in pop]
            muts = [agent.mut for agent in pop]

            banner_text = f"Global Steps {total_steps}"
            banner_width = max(len(banner_text) + 8, 35)
            border = "=" * banner_width
            centered_text = f"{banner_text}".center(banner_width)
            pbar.write(
                f"{border}\n"
                f"{centered_text}\n"
                f"{border}\n"
                f"Fitness:\t{fitness}\n"
                f"Score:\t\t{mean_scores}\n"
                f"5 fitness avgs:\t{avg_fitness}\n"
                f"10 score avgs:\t{avg_score}\n"
                f"Agents:\t\t{agents}\n"
                f"Steps:\t\t{num_steps}\n"
                f"Mutations:\t{muts}"
            )

        # Save model checkpoint
        if checkpoint is not None:
            if pop[0].steps[-1] // checkpoint > checkpoint_count:
                save_population_checkpoint(
                    population=pop,
                    save_path=save_path,
                    overwrite_checkpoints=overwrite_checkpoints,
                    accelerator=accelerator,
                )
                checkpoint_count += 1

    if wb:
        if accelerator is not None:
            accelerator.wait_for_everyone()
            if accelerator.is_main_process:
                wandb.finish()
            accelerator.wait_for_everyone()
        else:
            wandb.finish()

    pbar.close()
    return pop, pop_fitnesses



import random

import numpy as np
import pandas as pd

from agilerl.wrappers.learning import BanditEnv


# BanditEnv class creates an environment from a dataset
def test_create_environment():
    features = pd.DataFrame(np.random.uniform(0, 1, size=(10, 1)), columns=["features"])
    targets = pd.DataFrame(np.random.randint(0, 2, size=(10, 1)), columns=["targets"])

    env = BanditEnv(features, targets)  # Create environment
    context_dim = env.context_dim
    action_space = env.arms

    assert isinstance(env, BanditEnv)
    assert isinstance(action_space, int)
    assert isinstance(context_dim, tuple)
    assert action_space == int(targets.nunique()[0])
    assert context_dim == (len(np.array(features.loc[0])) * int(targets.nunique()),)


# BanditEnv class returns the reward output
def test_return_state_reward():
    features = pd.DataFrame(np.random.uniform(0, 1, size=(10, 1)), columns=["features"])
    targets = pd.DataFrame(np.random.randint(0, 2, size=(10, 1)), columns=["targets"])

    env = BanditEnv(features, targets)  # Create environment
    action_space = env.arms

    state = env.reset()
    action = random.randint(0, action_space - 1)
    prev_reward = env.prev_reward
    new_state, reward = env.step(action)

    assert state.shape == new_state.shape
    assert isinstance(reward, (int, float))
    assert reward == prev_reward[action]



import sys
from contextlib import contextmanager
from typing import Any
from unittest.mock import patch

import pytest
import torch
import torch.nn as nn
from accelerate import Accelerator
from accelerate.state import AcceleratorState
from datasets import Dataset as Datasets
from torch.utils.data import DataLoader, Dataset
from transformers import AutoTokenizer
from transformers.tokenization_utils_base import BatchEncoding

from agilerl.utils.algo_utils import DummyOptimizer
from agilerl.utils.llm_utils import (
    PreferenceGym,
    ReasoningGym,
    gather_if_zero3,
    get_state_dict,
)

DUMMY_CONVERSATION_TEMPLATE = [
    {
        "role": "system",
        "content": "question: {question}\nanswer: {answer}",
    },
]


class DummyTokenizer:
    def __init__(self, *args, **kwargs):
        self.vocab_size = 1000

    def batch_decode(self, *args, **kwargs):
        return ["This is a test completion."]

    def apply_chat_template(self, *args, **kwargs):
        return "This is a test completion."

    def __call__(self, *args, **kwargs):
        return torch.tensor([1, 2, 3, 4, 5])


class Info:
    def __init__(self, name):
        self.dataset_name = name


class DummyReasoningDataset(Dataset):
    def __init__(self, num_samples):
        # Create dummy questions and answers
        self.questions = [f"This is question {i}?" for i in range(num_samples)]
        self.answers = [f"This is answer {i}." for i in range(num_samples)]
        self.features = {"question": self.questions, "answer": self.answers}
        self.info = Info("dummy_dataset")

    def __len__(self):
        return len(self.questions)

    def __getitem__(self, index):
        return {"question": self.questions[index], "answer": self.answers[index]}


class DummyPreferenceDataset(Dataset):
    def __init__(self, num_samples):
        self.prompt = [f"This is prompt {i}." for i in range(num_samples)]
        self.chosen = [f"This is chosen {i}." for i in range(num_samples)]
        self.rejected = [f"This is rejected {i}." for i in range(num_samples)]
        self.features = {
            "prompt": self.prompt,
            "chosen": self.chosen,
            "rejected": self.rejected,
        }
        self.info = Info("dummy_dataset")

    def __len__(self):
        return len(self.prompt)

    def __getitem__(self, index):
        return {
            "prompt": self.prompt[index],
            "chosen": self.chosen[index],
            "rejected": self.rejected[index],
        }


def dummy_reward_fn(*args, **kwargs):
    return 1.0


def dummy_chat_template_fn_custom(q, a, tokenizer):
    """
    Chat template function for test_reasoning_gym_reset_dataloaders, gives unique input_ids for each question so
    we can test equality.
    """
    index = int(q.split(" ")[-1][0])
    return {
        "input_ids": torch.tensor([index]),
        "attention_mask": torch.ones(1),
    }


def dummy_chat_template_fn(q, a, tokenizer):
    return {
        "input_ids": torch.randint(0, 1000, (1, 356)),
        "attention_mask": torch.ones(1, 356),
    }


@pytest.fixture(scope="function")
def accelerator_factory():
    def generate_accelerator(use_accelerator):
        AcceleratorState._reset_state(True)
        return Accelerator() if use_accelerator else None

    return generate_accelerator


@pytest.fixture
def reasoning_dataset(num_samples):
    train_dataset = DummyReasoningDataset(int(num_samples * 0.8))
    test_dataset = DummyReasoningDataset(int(num_samples * 0.2))
    return train_dataset, test_dataset


@pytest.fixture
def preference_dataset(num_samples):
    train_dataset = DummyPreferenceDataset(int(num_samples * 0.8))
    test_dataset = DummyPreferenceDataset(int(num_samples * 0.2))
    return train_dataset, test_dataset


@pytest.mark.parametrize("num_samples", [200])
@pytest.mark.parametrize("use_accelerator", [True, False])
def test_reasoning_gym_init(
    reasoning_dataset, accelerator_factory, num_samples, use_accelerator
):
    train_dataset, test_dataset = reasoning_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 8
    env = ReasoningGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        reward_fn=dummy_reward_fn,
        conversation_template=DUMMY_CONVERSATION_TEMPLATE,
        data_batch_size_per_gpu=data_batch_size,
        accelerator=accelerator_factory(use_accelerator),
    )
    assert env.name == "dummy_dataset"
    assert callable(env.reward_fn)
    assert hasattr(env, "tokenizer")
    assert env.tokenizer is not None
    assert isinstance(env.train_dataloader, DataLoader)
    assert isinstance(env.test_dataloader, DataLoader)
    assert list(next(env.train_dataloader_iter).keys()) == [
        "question",
        "answer",
        "tokenized_prompts",
    ]
    assert env.dataloader == env.train_dataloader_iter
    assert not env.reset_called
    assert not env.evaluation_mode
    assert env.data_batch_size_per_gpu == data_batch_size


@pytest.mark.parametrize("num_samples", [200])
@pytest.mark.parametrize("eval_mode", [True, False])
@pytest.mark.parametrize("return_raw_completions", [True, False])
def test_reasoning_gym_step(
    reasoning_dataset, num_samples, eval_mode, return_raw_completions
):
    train_dataset, test_dataset = reasoning_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 8
    env = ReasoningGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        reward_fn=dummy_reward_fn,
        conversation_template=DUMMY_CONVERSATION_TEMPLATE,
        data_batch_size_per_gpu=data_batch_size,
        return_raw_completions=return_raw_completions,
    )
    env.evaluation_mode = eval_mode
    env.reset()
    completions = [torch.randint(0, 1000, (10, 356)) for _ in range(data_batch_size)]
    tokenized_prompts, rewards = env.step(completions)
    assert isinstance(tokenized_prompts, list)
    assert isinstance(rewards, torch.Tensor)

    for prompts in tokenized_prompts:
        assert sorted(prompts.keys()) == ["attention_mask", "input_ids", "text"]
        for key, val in prompts.items():
            match key:
                case "attention_mask":
                    assert isinstance(val, torch.Tensor)
                case "input_ids":
                    assert isinstance(val, torch.Tensor)
                case "text":
                    if return_raw_completions:
                        assert isinstance(val, str)
                    else:
                        assert val is None


@pytest.mark.parametrize("num_samples", [200])
@pytest.mark.parametrize("reset_dataloaders", [True, False])
@pytest.mark.parametrize("return_raw_completions", [True, False])
def test_reasoning_gym_reset(
    reasoning_dataset, num_samples, reset_dataloaders, return_raw_completions
):
    train_dataset, test_dataset = reasoning_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 8
    env = ReasoningGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        reward_fn=dummy_reward_fn,
        conversation_template=DUMMY_CONVERSATION_TEMPLATE,
        data_batch_size_per_gpu=data_batch_size,
        return_raw_completions=return_raw_completions,
    )
    tokenized_prompts = env.reset(reset_dataloaders)
    assert isinstance(tokenized_prompts, list)

    for prompts in tokenized_prompts:
        assert sorted(prompts.keys()) == ["attention_mask", "input_ids", "text"]
        for key, val in prompts.items():
            match key:
                case "attention_mask":
                    assert isinstance(val, torch.Tensor)
                case "input_ids":
                    assert isinstance(val, torch.Tensor)
                case "text":
                    if return_raw_completions:
                        assert isinstance(val, str)
                    else:
                        assert val is None


@pytest.mark.parametrize("num_samples", [200])
@pytest.mark.parametrize("reset_dataloaders", [True, False])
def test_reasoning_gym_reset_dataloaders(
    reasoning_dataset, num_samples, reset_dataloaders
):
    train_dataset, test_dataset = reasoning_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 8
    env = ReasoningGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        reward_fn=dummy_reward_fn,
        conversation_template=DUMMY_CONVERSATION_TEMPLATE,
        data_batch_size_per_gpu=data_batch_size,
    )
    first_data_point = next(
        env.test_dataloader_iter
    )  # use test_dataloader_iter as it is not shuffled
    env._reset_dataloaders()
    first_data_point_reset = next(env.test_dataloader_iter)
    for key1, key2 in zip(first_data_point.keys(), first_data_point_reset.keys()):
        if key1 == "tokenized_prompts":
            for item1, item2 in zip(
                first_data_point["tokenized_prompts"],
                first_data_point_reset["tokenized_prompts"],
            ):
                for key3, key4 in zip(item1.keys(), item2.keys()):
                    assert torch.equal(item1[key3], item2[key4])
        else:
            assert first_data_point[key1] == first_data_point_reset[key1]


@pytest.mark.parametrize("num_samples", [200])
def test_reset_warning(reasoning_dataset, num_samples):
    train_dataset, test_dataset = reasoning_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 8
    env = ReasoningGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        reward_fn=dummy_reward_fn,
        conversation_template=DUMMY_CONVERSATION_TEMPLATE,
        data_batch_size_per_gpu=data_batch_size,
    )
    with pytest.warns():
        env.reset()
        env.reset()


@pytest.mark.parametrize("num_samples", [200])
def test_reasoning_gym_len(reasoning_dataset, num_samples):
    train_dataset, test_dataset = reasoning_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 8
    env = ReasoningGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        reward_fn=dummy_reward_fn,
        conversation_template=DUMMY_CONVERSATION_TEMPLATE,
        data_batch_size_per_gpu=data_batch_size,
    )
    env.reset()
    assert len(env) == 200 * 0.8  # Length returns the training length
    with env.eval_mode():
        assert len(env) == 200 * 0.2


@pytest.mark.parametrize("num_samples", [20])
def test_create_chat_collate_fn(reasoning_dataset, num_samples):
    """Test the create_chat_collate_fn method."""
    # Create a mock tokenizer
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")

    train_dataset, test_dataset = reasoning_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 8

    env = ReasoningGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        reward_fn=dummy_reward_fn,
        conversation_template=DUMMY_CONVERSATION_TEMPLATE,
        data_batch_size_per_gpu=data_batch_size,
    )

    # Create the collate function
    collate_fn = env.create_collate_fn(tokenizer)

    # Create a sample batch
    batch = [
        {"question": "What is 2+2?", "answer": "4"},
        {"question": "What is 3+3?", "answer": "6"},
    ]

    # Apply the collate function
    result = collate_fn(batch)

    # Verify the result structure
    assert isinstance(result, dict)
    assert "question" in result
    assert "answer" in result
    assert "tokenized_prompts" in result

    # Verify the content
    assert result["question"] == ["What is 2+2?", "What is 3+3?"]
    assert result["answer"] == ["4", "6"]
    assert len(result["tokenized_prompts"]) == 2

    # Verify each tokenized prompt
    for prompt in result["tokenized_prompts"]:
        assert isinstance(prompt, BatchEncoding)
        assert "input_ids" in prompt
        assert "attention_mask" in prompt
        assert isinstance(prompt["input_ids"], torch.Tensor)
        assert isinstance(prompt["attention_mask"], torch.Tensor)


@pytest.mark.parametrize("num_samples", [20])
@pytest.mark.parametrize("data_batch_size", [8, 10])
def test_reset_dataloaders_when_train_dataloader_exhausted(
    reasoning_dataset, num_samples, data_batch_size
):
    train_dataset, test_dataset = reasoning_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    env = ReasoningGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        reward_fn=dummy_reward_fn,
        conversation_template=DUMMY_CONVERSATION_TEMPLATE,
        data_batch_size_per_gpu=data_batch_size,
    )
    total_sampled = 0
    for _ in range(3):
        env._get_next_batch()
        total_sampled += data_batch_size

    assert env.num_epochs == 1


@pytest.mark.parametrize("num_samples", [20])
@pytest.mark.parametrize("data_batch_size", [8, 10])
def test_not_reset_dataloaders_when_test_dataloader_exhausted(
    reasoning_dataset, num_samples, data_batch_size
):
    train_dataset, test_dataset = reasoning_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    env = ReasoningGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        reward_fn=dummy_reward_fn,
        conversation_template=DUMMY_CONVERSATION_TEMPLATE,
        data_batch_size_per_gpu=data_batch_size,
    )
    total_sampled = 0
    env.reset()
    for _ in range(10):
        with env.eval_mode():
            env._get_next_batch()
            total_sampled += data_batch_size

    assert env.num_epochs == 0


def test_dummy_optimizer_init():
    """Test DummyOptimizer initialization."""
    params = [torch.tensor([1.0, 2.0, 3.0])]
    lr = 0.001
    optimizer = DummyOptimizer(params, lr)
    assert optimizer is not None


def test_dummy_optimizer_step():
    """Test DummyOptimizer step method raises RuntimeError."""
    params = [torch.tensor([1.0, 2.0, 3.0])]
    lr = 0.001
    optimizer = DummyOptimizer(params, lr)

    with pytest.raises(RuntimeError) as exc_info:
        optimizer.step()

    expected_message = (
        "DummyOptimizer is a placeholder optimizer and should not be used."
        "Please ensure you are calling accelerator.prepare() on the optimizer."
    )
    assert str(exc_info.value) == expected_message


def test_dummy_optimizer_zero_grad():
    """Test DummyOptimizer zero_grad method raises RuntimeError."""
    params = [torch.tensor([1.0, 2.0, 3.0])]
    lr = 0.001
    optimizer = DummyOptimizer(params, lr)

    with pytest.raises(RuntimeError) as exc_info:
        optimizer.zero_grad()

    expected_message = (
        "DummyOptimizer is a placeholder optimizer and should not be used."
        "Please ensure you are calling accelerator.prepare() on the optimizer."
    )
    assert str(exc_info.value) == expected_message


def test_dummy_optimizer_state_dict():
    """Test DummyOptimizer state_dict method raises RuntimeError."""
    params = [torch.tensor([1.0, 2.0, 3.0])]
    lr = 0.001
    optimizer = DummyOptimizer(params, lr)

    with pytest.raises(RuntimeError) as exc_info:
        optimizer.state_dict()

    expected_message = (
        "DummyOptimizer is a placeholder optimizer and should not be used."
        "Please ensure you are calling accelerator.prepare() on the optimizer."
    )
    assert str(exc_info.value) == expected_message


def test_dummy_optimizer_load_state_dict():
    """Test DummyOptimizer load_state_dict method raises RuntimeError."""
    params = [torch.tensor([1.0, 2.0, 3.0])]
    lr = 0.001
    optimizer = DummyOptimizer(params, lr)

    with pytest.raises(RuntimeError) as exc_info:
        optimizer.load_state_dict({})

    expected_message = (
        "DummyOptimizer is a placeholder optimizer and should not be used."
        "Please ensure you are calling accelerator.prepare() on the optimizer."
    )
    assert str(exc_info.value) == expected_message


@pytest.mark.parametrize("zero_stage", [0, 1, 2, 3])
def test_gather_if_zero3(zero_stage):
    """Test gather_if_zero3 context manager."""
    params = [torch.tensor([1.0, 2.0, 3.0])]

    @contextmanager
    def dummy_gather_parameters(*args, **kwargs):
        yield

    with patch(
        "deepspeed.zero.GatheredParameters", side_effect=dummy_gather_parameters
    ) as mock_gathered_parameters:
        with gather_if_zero3(zero_stage, params):
            assert mock_gathered_parameters.call_count == (zero_stage == 3)


@pytest.mark.parametrize("use_accelerator", [True, False])
@pytest.mark.parametrize("num_samples", [20])
def test_preference_gym_init(
    preference_dataset, accelerator_factory, use_accelerator, num_samples
):
    train_dataset, test_dataset = preference_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 8
    env = PreferenceGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        data_batch_size_per_gpu=data_batch_size,
        accelerator=accelerator_factory(use_accelerator),
    )
    assert env.name == "dummy_dataset"
    assert hasattr(env, "tokenizer")
    assert isinstance(env.train_dataloader, DataLoader)
    assert isinstance(env.test_dataloader, DataLoader)
    print("KEYS", next(env.train_dataloader_iter).keys())
    assert list(next(env.train_dataloader_iter).keys()) == [
        "prompt",
        "prompt_lengths",
        "chosen",
        "rejected",
        "chosen_input_ids",
        "chosen_attention_mask",
        "rejected_input_ids",
        "rejected_attention_mask",
    ]
    assert env.dataloader == env.train_dataloader_iter
    assert not env.reset_called
    assert not env.evaluation_mode
    assert env.data_batch_size_per_gpu == data_batch_size


@pytest.mark.parametrize("use_accelerator", [True, False])
@pytest.mark.parametrize("num_samples", [20])
def test_preference_gym_step(
    preference_dataset, accelerator_factory, use_accelerator, num_samples
):
    train_dataset, test_dataset = preference_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 8
    env = PreferenceGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        data_batch_size_per_gpu=data_batch_size,
        accelerator=accelerator_factory(use_accelerator),
    )
    prompts = env.step()
    assert isinstance(prompts, dict)
    assert set(prompts.keys()) == {
        "prompt",
        "prompt_lengths",
        "chosen",
        "rejected",
        "chosen_input_ids",
        "chosen_attention_mask",
        "rejected_input_ids",
        "rejected_attention_mask",
    }
    assert len(prompts["prompt"]) == data_batch_size
    assert len(prompts["prompt_lengths"]) == data_batch_size
    assert len(prompts["chosen"]) == data_batch_size
    assert len(prompts["rejected"]) == data_batch_size
    assert len(prompts["chosen_input_ids"]) == data_batch_size
    assert len(prompts["chosen_attention_mask"]) == data_batch_size
    assert len(prompts["rejected_input_ids"]) == data_batch_size
    assert len(prompts["rejected_attention_mask"]) == data_batch_size
    assert isinstance(prompts["prompt"], list)
    assert isinstance(prompts["prompt"][0], str)
    assert isinstance(prompts["prompt_lengths"][0], int)
    assert isinstance(prompts["prompt_lengths"], list)
    assert isinstance(prompts["chosen"], list)
    assert isinstance(prompts["rejected"], list)
    assert isinstance(prompts["chosen_input_ids"], torch.Tensor)
    assert isinstance(prompts["chosen_attention_mask"], torch.Tensor)
    assert isinstance(prompts["rejected_input_ids"], torch.Tensor)
    assert isinstance(prompts["rejected_attention_mask"], torch.Tensor)
    assert not env.reset_called


@pytest.mark.parametrize("use_accelerator", [True, False])
@pytest.mark.parametrize("num_samples", [20])
def test_preference_gym_reset(
    preference_dataset, accelerator_factory, use_accelerator, num_samples
):
    train_dataset, test_dataset = preference_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 8
    env = PreferenceGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        data_batch_size_per_gpu=data_batch_size,
        accelerator=accelerator_factory(use_accelerator),
    )
    prompts = env.reset()
    assert isinstance(prompts, dict)
    assert set(prompts.keys()) == {
        "prompt",
        "prompt_lengths",
        "chosen",
        "rejected",
        "chosen_input_ids",
        "chosen_attention_mask",
        "rejected_input_ids",
        "rejected_attention_mask",
    }
    assert len(prompts["prompt"]) == data_batch_size
    assert len(prompts["prompt_lengths"]) == data_batch_size
    assert len(prompts["chosen"]) == data_batch_size
    assert len(prompts["rejected"]) == data_batch_size
    assert len(prompts["chosen_input_ids"]) == data_batch_size
    assert len(prompts["chosen_attention_mask"]) == data_batch_size
    assert len(prompts["rejected_input_ids"]) == data_batch_size
    assert len(prompts["rejected_attention_mask"]) == data_batch_size
    assert isinstance(prompts["prompt"], list)
    assert isinstance(prompts["prompt"][0], str)
    assert isinstance(prompts["prompt_lengths"][0], int)
    assert isinstance(prompts["prompt_lengths"], list)
    assert isinstance(prompts["chosen"], list)
    assert isinstance(prompts["rejected"], list)
    assert isinstance(prompts["chosen_input_ids"], torch.Tensor)
    assert isinstance(prompts["chosen_attention_mask"], torch.Tensor)
    assert isinstance(prompts["rejected_input_ids"], torch.Tensor)
    assert isinstance(prompts["rejected_attention_mask"], torch.Tensor)
    assert env.reset_called


@pytest.mark.parametrize("use_accelerator", [True, False])
@pytest.mark.parametrize("num_samples", [20])
def test_preference_gym_reset_reset_dataloaders_warning(
    preference_dataset, accelerator_factory, use_accelerator, num_samples
):
    train_dataset, test_dataset = preference_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 1
    env = PreferenceGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        data_batch_size_per_gpu=data_batch_size,
        accelerator=accelerator_factory(use_accelerator),
    )
    env.reset()
    env.step()
    env.step()
    with pytest.warns(
        UserWarning,
        match=r"env\.reset\(\) called with reset_dataloaders=True, this will reset the dataloaders to the beginning of the dataset, proceed with caution\.",
    ):
        prompts = env.reset(reset_dataloaders=True)
    assert len(prompts["prompt"]) == data_batch_size
    assert isinstance(prompts, dict)
    assert set(prompts.keys()) == {
        "prompt",
        "prompt_lengths",
        "chosen",
        "rejected",
        "chosen_input_ids",
        "chosen_attention_mask",
        "rejected_input_ids",
        "rejected_attention_mask",
    }
    assert len(prompts["prompt"]) == data_batch_size
    assert len(prompts["prompt_lengths"]) == data_batch_size
    assert len(prompts["chosen"]) == data_batch_size
    assert len(prompts["rejected"]) == data_batch_size
    assert len(prompts["chosen_input_ids"]) == data_batch_size
    assert len(prompts["chosen_attention_mask"]) == data_batch_size
    assert len(prompts["rejected_input_ids"]) == data_batch_size
    assert len(prompts["rejected_attention_mask"]) == data_batch_size
    assert isinstance(prompts["prompt"], list)
    assert isinstance(prompts["prompt"][0], str)
    assert isinstance(prompts["prompt_lengths"][0], int)
    assert isinstance(prompts["prompt_lengths"], list)
    assert isinstance(prompts["chosen"], list)
    assert isinstance(prompts["rejected"], list)
    assert isinstance(prompts["chosen_input_ids"], torch.Tensor)
    assert isinstance(prompts["chosen_attention_mask"], torch.Tensor)
    assert isinstance(prompts["rejected_input_ids"], torch.Tensor)
    assert isinstance(prompts["rejected_attention_mask"], torch.Tensor)
    assert env.reset_called


@pytest.mark.parametrize("use_accelerator", [True, False])
@pytest.mark.parametrize("num_samples", [20])
def test_preference_gym_reset_reset_called_warning(
    preference_dataset, accelerator_factory, use_accelerator, num_samples
):
    train_dataset, test_dataset = preference_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 1
    env = PreferenceGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        data_batch_size_per_gpu=data_batch_size,
        accelerator=accelerator_factory(use_accelerator),
    )
    with pytest.warns(
        UserWarning,
        match=r"env\.reset\(\) called more than once sequentially, it should typically follow with env\.step\(\)\.",
    ):
        env.reset_called = True
        prompts = env.reset()
    assert len(prompts["prompt"]) == data_batch_size
    assert isinstance(prompts, dict)
    assert set(prompts.keys()) == {
        "prompt",
        "prompt_lengths",
        "chosen",
        "rejected",
        "chosen_input_ids",
        "chosen_attention_mask",
        "rejected_input_ids",
        "rejected_attention_mask",
    }
    assert len(prompts["prompt"]) == data_batch_size
    assert len(prompts["prompt_lengths"]) == data_batch_size
    assert len(prompts["chosen"]) == data_batch_size
    assert len(prompts["rejected"]) == data_batch_size
    assert len(prompts["chosen_input_ids"]) == data_batch_size
    assert len(prompts["chosen_attention_mask"]) == data_batch_size
    assert len(prompts["rejected_input_ids"]) == data_batch_size
    assert len(prompts["rejected_attention_mask"]) == data_batch_size
    assert isinstance(prompts["prompt"], list)
    assert isinstance(prompts["prompt"][0], str)
    assert isinstance(prompts["prompt_lengths"][0], int)
    assert isinstance(prompts["prompt_lengths"], list)
    assert isinstance(prompts["chosen"], list)
    assert isinstance(prompts["rejected"], list)
    assert isinstance(prompts["chosen_input_ids"], torch.Tensor)
    assert isinstance(prompts["chosen_attention_mask"], torch.Tensor)
    assert isinstance(prompts["rejected_input_ids"], torch.Tensor)
    assert isinstance(prompts["rejected_attention_mask"], torch.Tensor)
    assert env.reset_called


@pytest.mark.parametrize("num_samples", [20])
@pytest.mark.parametrize("use_accelerator", [True, False])
def test_preference_gym_reset_num_epochs(
    preference_dataset, num_samples, accelerator_factory, use_accelerator
):
    train_dataset, test_dataset = preference_dataset
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 1
    env = PreferenceGym(
        train_dataset=train_dataset,
        test_dataset=test_dataset,
        tokenizer=tokenizer,
        data_batch_size_per_gpu=data_batch_size,
        accelerator=accelerator_factory(use_accelerator),
    )
    while env.num_epochs == 0:
        env.step()
    assert env.num_epochs == 1


def test_get_state_dict():
    model = nn.Linear(10, 10)
    state_dict = get_state_dict(model)
    assert isinstance(state_dict, dict)
    for key, value in state_dict.items():
        assert isinstance(key, str)
        assert isinstance(value, torch.Tensor)


def test_preference_gym_max_context_length_error():
    train_dataset = Datasets.from_dict(
        {
            "prompt": ["This is a prompt that is longer than the max context length."],
            "chosen": ["This is an answer."],
            "rejected": ["This is an answer."],
        }
    )
    test_dataset = Datasets.from_dict(
        {
            "prompt": ["This is a normal length prompt"],
            "chosen": ["This is an answer."],
            "rejected": ["This is an answer."],
        }
    )
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 8
    with pytest.raises(
        ValueError,
        match="No samples left in the train dataset after filtering by the max context length constraint, use a larger max context length.",
    ):
        PreferenceGym(
            train_dataset=train_dataset,
            test_dataset=test_dataset,
            tokenizer=tokenizer,
            data_batch_size_per_gpu=data_batch_size,
            max_context_length=5,
            min_completion_length=1,
        )


def test_preference_gym_max_context_length_warning():
    train_dataset = Datasets.from_dict(
        {
            "prompt": [
                "This is a prompt that is longer than the max context length. This prompt really is a lot longer than the other one.",
                "This is a prompt that is shorter.",
            ],
            "chosen": ["This is an answer.", "This is an answer."],
            "rejected": ["This is an answer.", "This is an answer."],
        }
    )
    test_dataset = Datasets.from_dict(
        {
            "prompt": ["This is a normal length prompt"],
            "chosen": ["This is an answer."],
            "rejected": ["This is an answer."],
        }
    )
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 8
    with pytest.warns(
        UserWarning,
        match=r"1 samples were filtered out of the train dataset due to the max context length constraint.",
    ):
        env = PreferenceGym(
            train_dataset=train_dataset,
            test_dataset=test_dataset,
            tokenizer=tokenizer,
            data_batch_size_per_gpu=data_batch_size,
            max_context_length=10,
            min_completion_length=1,
        )
    assert len(env.train_dataloader) == 1
    assert len(env.test_dataloader) == 1


def test_reasoning_gym_max_context_length_warning():
    train_dataset = Datasets.from_dict(
        {
            "question": [
                "This is a prompt that is longer than the max context length. This prompt really is a lot longer than the other one.",
                "This is a prompt that is shorter.",
            ],
            "answer": ["This is an answer.", "This is an answer."],
        }
    )
    test_dataset = Datasets.from_dict(
        {
            "question": ["This is a normal length prompt"],
            "answer": ["This is an answer."],
        }
    )
    tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-0.5B")
    data_batch_size = 8
    with pytest.warns(
        UserWarning,
        match=r"1 samples were filtered out of the train dataset due to the max context length constraint.",
    ):
        env = ReasoningGym(
            train_dataset=train_dataset,
            test_dataset=test_dataset,
            tokenizer=tokenizer,
            reward_fn=dummy_reward_fn,
            conversation_template=DUMMY_CONVERSATION_TEMPLATE,
            data_batch_size_per_gpu=data_batch_size,
            max_context_length=10,
        )
    assert len(env.train_dataloader) == 1
    assert len(env.test_dataloader) == 1


def test_llm_utils_fallback_types_when_no_llm_dependencies():
    """Test that llm_utils sets type aliases to Any when HAS_LLM_DEPENDENCIES is False."""
    # Remove the module from cache to force reimport
    original_module = sys.modules.pop("agilerl.utils.llm_utils", None)

    try:
        # Patch HAS_LLM_DEPENDENCIES before reimporting
        with patch("agilerl.HAS_LLM_DEPENDENCIES", False):
            # Reimport the module - it will see HAS_LLM_DEPENDENCIES as False
            import agilerl.utils.llm_utils as llm_utils_reloaded

            # Verify the fallback type aliases are set to Any
            assert llm_utils_reloaded.AutoTokenizer is Any
            assert llm_utils_reloaded.PreTrainedModel is Any
            assert llm_utils_reloaded.BatchEncoding is Any
            assert llm_utils_reloaded.Dataset is Any
    finally:
        # Restore original module to avoid affecting other tests
        sys.modules["agilerl.utils.llm_utils"] = original_module



import warnings
from typing import Any, Optional, Union

import torch
from tensordict import TensorDict
from torch.utils.data import DataLoader

from agilerl.components import (
    MultiAgentReplayBuffer,
    MultiStepReplayBuffer,
    PrioritizedReplayBuffer,
    ReplayBuffer,
)
from agilerl.components.data import ReplayDataset
from agilerl.typing import ExperiencesType

BufferType = Union[
    ReplayBuffer, MultiAgentReplayBuffer, PrioritizedReplayBuffer, MultiStepReplayBuffer
]


class Sampler:
    """Sampler class to handle both standard and distributed training.

    :param memory: Replay buffer memory, defaults to None
    :type memory: Optional[Union[ReplayBuffer, MultiAgentReplayBuffer, PrioritizedReplayBuffer, MultiStepReplayBuffer]], optional
    :param dataset: Dataset for distributed sampling, defaults to None
    :type dataset: Optional[ReplayDataset], optional
    :param dataloader: DataLoader for distributed sampling, defaults to None
    :type dataloader: Optional[DataLoader], optional
    :raises AssertionError: If neither memory nor (dataset and dataloader) are provided
    """

    @staticmethod
    def tensordict_collate_fn(
        batch: list[TensorDict],
    ) -> Union[TensorDict, list[TensorDict]]:
        """Custom collate function that properly handles TensorDict objects.

        :param batch: List of TensorDict objects to collate
        :type batch: list[TensorDict]
        :return: Either a single TensorDict or a list of TensorDicts
        :rtype: Union[TensorDict, list[TensorDict]]
        """
        return TensorDict(
            {key: torch.stack([b[key] for b in batch]) for key in batch[0].keys()},
            batch_size=len(batch),
        )

    def __init__(
        self,
        memory: Optional[BufferType] = None,
        dataset: Optional[ReplayDataset] = None,
        dataloader: Optional[DataLoader] = None,
    ) -> None:

        assert (memory is not None) or (
            (dataset is not None) and (dataloader is not None)
        ), "Sampler needs to be initialized with either 'memory' or ('dataset' AND 'dataloader')."

        self.distributed = (
            dataloader is not None
            and dataset is not None
            and isinstance(dataloader, DataLoader)
        )
        self.per = isinstance(memory, PrioritizedReplayBuffer)
        self.n_step = isinstance(memory, MultiStepReplayBuffer)
        self.memory = memory
        self.dataset = dataset

        # Process the dataloader
        if self.distributed:
            # Need to use a custom collate function for TensorDict buffers
            if isinstance(dataset.buffer, ReplayBuffer):
                self.dataloader = self._replace_dataloader_collate_fn(dataloader)
            else:
                self.dataloader = dataloader

            if not isinstance(self.dataset, ReplayDataset):
                warnings.warn("Dataset is not an agilerl ReplayDataset.")
            if not isinstance(self.dataloader, DataLoader):
                warnings.warn("Dataset is not a torch DataLoader object.")

            self.sample = self.sample_distributed
        else:
            self.dataloader = dataloader
            if self.per:
                if not isinstance(self.memory, PrioritizedReplayBuffer):
                    warnings.warn("Memory is not an agilerl PrioritizedReplayBuffer.")
                self.sample = self.sample_per

            elif self.n_step:
                if not isinstance(self.memory, MultiStepReplayBuffer):
                    warnings.warn("Memory is not an agilerl MultiStepReplayBuffer.")
                self.sample = self.sample_n_step

            else:
                if not isinstance(self.memory, (ReplayBuffer, MultiAgentReplayBuffer)):
                    warnings.warn(
                        "Memory is not an agilerl ReplayBuffer or MultiAgentReplayBuffer."
                    )
                self.sample = self.sample_standard

    def _replace_dataloader_collate_fn(self, dataloader: DataLoader) -> DataLoader:
        """Create a new DataLoader with the tensordict_collate_fn while preserving all other parameters.

        :param dataloader: Original DataLoader
        :type dataloader: DataLoader
        :return: New DataLoader with tensordict_collate_fn
        :rtype: DataLoader
        """
        # Create a simplified set of parameters for the new dataloader
        params = {
            "dataset": dataloader.dataset,
            "batch_size": dataloader.batch_size,
            "collate_fn": self.tensordict_collate_fn,
        }

        # Optional parameters - only include if they are not default values
        if dataloader.num_workers != 0:
            params["num_workers"] = dataloader.num_workers

        if dataloader.pin_memory:
            params["pin_memory"] = True

        if dataloader.drop_last:
            params["drop_last"] = True

        if hasattr(dataloader, "prefetch_factor") and dataloader.prefetch_factor != 2:
            params["prefetch_factor"] = dataloader.prefetch_factor

        if hasattr(dataloader, "persistent_workers") and dataloader.persistent_workers:
            params["persistent_workers"] = True

        # Create a new DataLoader
        return DataLoader(**params)

    def sample_standard(
        self, batch_size: int, return_idx: bool = False
    ) -> ExperiencesType:
        """Sample a batch of experiences from the standard replay buffer.

        :param batch_size: Size of the batch to sample
        :type batch_size: int
        :param return_idx: Whether to return indices, defaults to False
        :type return_idx: bool, optional
        :return: Sampled batch of experiences
        :rtype: TensorDict
        """
        return self.memory.sample(batch_size, return_idx)

    def sample_distributed(
        self, batch_size: int, return_idx: Optional[bool] = None
    ) -> TensorDict:
        """Sample a batch of experiences from the distributed dataset.

        :param batch_size: Size of the batch to sample
        :type batch_size: int
        :param return_idx: Not used in distributed sampling, defaults to None
        :type return_idx: Optional[bool], optional
        :return: Sampled batch of experiences
        :rtype: TensorDict
        """
        self.dataset.batch_size = batch_size
        return next(iter(self.dataloader))

    def sample_per(self, batch_size: int, beta: float) -> ExperiencesType:
        """Sample a batch of experiences from the Prioritized Experience Replay buffer.

        :param batch_size: Size of the batch to sample
        :type batch_size: int
        :param beta: Importance-sampling weight
        :type beta: float
        :return: Sampled batch of experiences, indices, and importance-sampling weights
        :rtype: TensorDict
        """
        return self.memory.sample(batch_size, beta)

    def sample_n_step(self, idxs: Any) -> ExperiencesType:
        """Sample a batch of experiences from the n-step replay buffer.

        :param idxs: Indices to sample from
        :type idxs: Any
        :return: Sampled batch of experiences
        :rtype: TensorDict
        """
        return self.memory.sample_from_indices(idxs)

    @classmethod
    def create_dataloader(
        cls, dataset: ReplayDataset, batch_size: Optional[int] = None, **kwargs
    ) -> DataLoader:
        """Helper method to create a DataLoader with the appropriate collate function.

        :param dataset: Dataset to create a DataLoader for
        :type dataset: ReplayDataset
        :param batch_size: Batch size for the DataLoader, defaults to None
        :type batch_size: Optional[int], optional
        :param kwargs: Additional arguments to pass to the DataLoader
        :return: DataLoader with tensordict_collate_fn
        :rtype: DataLoader
        """
        # Ensure we don't override collate_fn if explicitly provided
        if "collate_fn" not in kwargs:
            kwargs["collate_fn"] = cls.tensordict_collate_fn

        return DataLoader(dataset, batch_size=batch_size, **kwargs)



import gc
import random

import deepspeed.comm.comm as ds_comm
import deepspeed.utils.groups as ds_groups
import numpy as np
import pytest
import torch
from accelerate import Accelerator
from accelerate.state import AcceleratorState
from accelerate.utils import DeepSpeedPlugin
from peft import LoraConfig, get_peft_model
from transformers import AutoModelForCausalLM
from vllm.distributed import (
    cleanup_dist_env_and_memory,
)


@pytest.fixture(autouse=True)
def cleanup_after_test(request):
    yield
    if "vllm" in request.node.name:
        cleanup_dist_env_and_memory()
        for attr in dir(ds_groups):
            if attr.startswith("_") and attr.endswith("_GROUP"):
                setattr(ds_groups, attr, None)
        ds_comm.cdb = None
    gc.collect()
    torch.cuda.synchronize()
    torch.cuda.empty_cache()
    AcceleratorState._reset_state(True)


@pytest.fixture(autouse=True)
def set_seed():
    """Set random seeds for reproducibility."""
    SEED = 42
    torch.manual_seed(SEED)
    torch.cuda.manual_seed(SEED)
    torch.cuda.manual_seed_all(SEED)
    np.random.seed(SEED)
    random.seed(SEED)
    yield


@pytest.fixture(scope="function")
def accelerator_factory():
    def generate_accelerator(use_deepspeed_optimizer, config):
        gc.collect()
        torch.cuda.empty_cache()
        AcceleratorState._reset_state(True)
        if use_deepspeed_optimizer and (config is not None):
            config["optimizer"] = {
                "type": "AdamW",
                "params": {
                    "lr": 1e-4,  # Smaller learning rate
                    "betas": [0.9, 0.999],
                    "eps": 1e-8,
                    "weight_decay": 0.01,
                },
            }
        return (
            Accelerator(deepspeed_plugin=DeepSpeedPlugin(hf_ds_config=config))
            if config is not None
            else None
        )

    return generate_accelerator


@pytest.fixture(scope="function")
def model_factory():
    def generate_model(pretrained_model_name_or_path):
        peft_config = LoraConfig(
            task_type="CAUSAL_LM",
            r=16,
            lora_alpha=64,
            target_modules=[
                "q_proj",
                "k_proj",
                "v_proj",
                "o_proj",
                "gate_proj",
                "up_proj",
                "down_proj",
            ],
        )
        model = AutoModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            torch_dtype=torch.bfloat16,
            attn_implementation="sdpa",
        )
        model.gradient_checkpointing_enable()
        model = get_peft_model(model, peft_config)
        return model

    return generate_model



"""Functions for collecting rollouts for on-policy algorithms."""

from typing import Any, Optional

import numpy as np
import torch
from gymnasium import spaces

from agilerl.algorithms import PPO
from agilerl.networks import StochasticActor
from agilerl.typing import GymEnvType

SupportedOnPolicy = PPO


def _collect_rollouts(
    agent: SupportedOnPolicy,
    env: GymEnvType,
    n_steps: Optional[int] = None,
    last_obs: Optional[np.ndarray] = None,
    last_done: Optional[np.ndarray] = None,
    last_scores: Optional[np.ndarray] = None,
    last_info: Optional[dict[str, Any]] = None,
    *,
    recurrent: bool,
) -> tuple[list[float], np.ndarray, np.ndarray, np.ndarray, dict[str, Any]]:
    """Collect rollouts for on-policy algorithms.

    :param agent: The agent to collect rollouts for.
    :type agent: SupportedOnPolicy
    :param env: The environment to collect rollouts from.
    :type env: GymEnvType
    :param n_steps: The number of steps to collect rollouts for. Defaults to agent.learn_step if not provided.
    :type n_steps: Optional[int]
    :param last_obs: The observation to use for the first step. Defaults to None, where the environment is reset.
    :type last_obs: Optional[np.ndarray]
    :param last_done: The done flag to use for the first step. Defaults to None, where the environment is reset.
    :type last_done: Optional[np.ndarray]
    :param last_scores: The scores to use for the first step. Defaults to None, where the environment is reset.
    :type last_scores: Optional[np.ndarray]
    :param last_info: The info for the current step. Defaults to None, where the environment is reset.
    :type last_info: Optional[dict[str, Any]]
    :param recurrent: Whether the agent is recurrent.
    :type recurrent: bool

    :return: The list of scores for the episodes completed in the rollouts
    :rtype: list[float]
    :return: The observation, done flag, scores, and info for the current step.
    :rtype: tuple[np.ndarray, np.ndarray, np.ndarray, dict[str, Any]]
    """
    if not agent.use_rollout_buffer:
        raise RuntimeError(
            "collect_rollouts can only be used when use_rollout_buffer=True"
        )

    if (
        last_obs is None
        and last_done is None
        and last_scores is None
        and last_info is None
    ):
        obs, info = env.reset()
        scores = np.zeros(agent.num_envs)
        done = np.zeros(agent.num_envs)
        agent.hidden_state = (
            agent.get_initial_hidden_state(agent.num_envs) if recurrent else None
        )
    else:
        obs = last_obs
        done = last_done
        scores = last_scores
        info = last_info

    n_steps = n_steps or -(agent.learn_step // -agent.num_envs)
    agent.rollout_buffer.reset()

    current_hidden_state_for_actor = agent.hidden_state

    completed_episode_scores = []
    for _ in range(n_steps):
        current_hidden_state_for_buffer = current_hidden_state_for_actor

        # Get action, statistics and (maybe) recurrent hidden state from agent
        if recurrent:
            action, log_prob, _, value, next_hidden_for_actor = agent.get_action(
                obs,
                action_mask=info.get("action_mask", None),
                hidden_state=current_hidden_state_for_actor,
            )
            agent.hidden_state = next_hidden_for_actor
        else:
            action, log_prob, _, value = agent.get_action(
                obs, action_mask=info.get("action_mask", None)
            )

        # Clip action to action space
        policy = getattr(agent, agent.registry.policy())
        if isinstance(policy, StochasticActor) and isinstance(
            agent.action_space, spaces.Box
        ):
            if policy.squash_output:
                clipped_action = policy.scale_action(action)
            else:
                clipped_action = np.clip(
                    action,
                    agent.action_space.low,
                    agent.action_space.high,
                )
        else:
            clipped_action = action

        next_obs, reward, term, trunc, next_info = env.step(clipped_action)

        # Check if termination condition is met
        if isinstance(term, (list, np.ndarray)):
            is_terminal = (
                np.logical_or(term, trunc)
                if isinstance(trunc, (list, np.ndarray))
                else term
            )
        else:
            is_terminal = term or trunc

        reward_np = np.atleast_1d(reward)
        is_terminal_np = np.atleast_1d(is_terminal)
        value_np = np.atleast_1d(value)
        log_prob_np = np.atleast_1d(log_prob)

        agent.rollout_buffer.add(
            obs=obs,
            action=action,
            reward=reward_np,
            done=is_terminal_np,
            value=value_np,
            log_prob=log_prob_np,
            next_obs=next_obs,
            hidden_state=current_hidden_state_for_buffer,
        )

        scores += reward_np
        done = is_terminal_np

        if recurrent and np.any(is_terminal_np):
            finished_mask = is_terminal_np.astype(bool)
            initial_hidden_states_for_reset = agent.get_initial_hidden_state(
                agent.num_envs
            )
            if isinstance(agent.hidden_state, dict):
                for key in agent.hidden_state:
                    reset_states_for_key = initial_hidden_states_for_reset[key][
                        :, finished_mask, :
                    ]
                    if reset_states_for_key.shape[1] > 0:
                        agent.hidden_state[key][
                            :, finished_mask, :
                        ] = reset_states_for_key

        if recurrent:
            current_hidden_state_for_actor = agent.hidden_state

        obs = next_obs
        info = next_info

        for idx, env_done in enumerate(is_terminal_np):
            if env_done:
                completed_episode_scores.append(scores[idx])
                agent.scores.append(scores[idx])
                scores[idx] = 0

    # Calculate last value to compute returns and advantages properly
    with torch.no_grad():
        if recurrent:
            _, _, _, last_value, _ = agent._get_action_and_values(
                agent.preprocess_observation(obs),
                hidden_state=agent.hidden_state,
            )
        else:
            _, _, _, last_value, _ = agent._get_action_and_values(
                agent.preprocess_observation(obs)
            )

        last_value = last_value.cpu().numpy()
        last_done = np.atleast_1d(term)

    agent.rollout_buffer.compute_returns_and_advantages(
        last_value=last_value, last_done=last_done
    )

    return completed_episode_scores, obs, done, scores, info


def collect_rollouts(
    agent: SupportedOnPolicy, env: GymEnvType, n_steps: Optional[int] = None, **kwargs
) -> list[float]:
    """Collect rollouts for non-recurrent on-policy algorithms.

    :param agent: The agent to collect rollouts for.
    :type agent: RLAlgorithm
    :param env: The environment to collect rollouts from.
    :type env: GymEnvType
    :param n_steps: The number of steps to collect rollouts for.
    :type n_steps: Optional[int]

    :return: The list of scores for the episodes completed in the rollouts
    :rtype: list[float]
    """

    return _collect_rollouts(agent, env, n_steps, recurrent=False, **kwargs)


def collect_rollouts_recurrent(
    agent: SupportedOnPolicy, env: GymEnvType, n_steps: Optional[int] = None, **kwargs
) -> list[float]:
    """Collect rollouts for recurrent on-policy algorithms.

    :param agent: The agent to collect rollouts for.
    :type agent: RLAlgorithm
    :param env: The environment to collect rollouts from.
    :type env: GymEnvType
    :param n_steps: The number of steps to collect rollouts for.
    :type n_steps: Optional[int]

    :return: The list of scores for the episodes completed in the rollouts
    :rtype: list[float]
    """

    return _collect_rollouts(agent, env, n_steps, recurrent=True, **kwargs)



import copy

import numpy as np
import pytest
import torch

from agilerl.modules.cnn import EvolvableCNN
from tests.helper_functions import assert_state_dicts_equal


######### Test instantiation #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs",
    [
        ([1, 16, 16], [32], [3], [1], 10),
        ([1, 16, 16], [32], [(3, 3)], [(1, 1)], 10),
        ([3, 128, 128], [8, 8, 8], [2, 2, 2], [2, 2, 2], 1),
        ([1, 64], [32], [3], [1], 10),  # Conv1D case
    ],
)
def test_instantiation_without_errors(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    device,
):
    block_type = "Conv1d" if len(input_shape) == 2 else "Conv2d"  # Infer block type
    evolvable_cnn = EvolvableCNN(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        block_type=block_type,
        device=device,
    )
    assert isinstance(evolvable_cnn, EvolvableCNN)


@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs, explicit_block_type",
    [
        (
            [1, 16, 16],
            [32],
            [3, 3],
            [1],
            10,
            None,
        ),  # kernel vs channel len mismatch for Conv2d
        (
            [1, 16, 16],
            [32],
            [3],
            [1, 1],
            10,
            None,
        ),  # stride vs channel len mismatch for Conv2d
        ([1, 16, 16], [32], [(3, 3)], [(1, 1)], 0, None),  # num_outputs <= 0 for Conv2d
        (
            [3, 128],
            [8, 8, 8],
            [2, 2, 2],
            [2, 2, 2],
            1,
            "Conv2d",
        ),  # Invalid input_shape dim for Conv2d
        ([1], [32], [3], [1], 10, "Conv1d"),  # Conv1D incorrect input_shape dim
        (
            [1, 64],
            [32],
            [3, 3],
            [1],
            10,
            "Conv1d",
        ),  # Conv1D kernel_size mismatch with channel_size
    ],
)
def test_incorrect_instantiation(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    explicit_block_type,  # New parameter
    device,
):
    if explicit_block_type:
        block_type = explicit_block_type
    else:
        # Default inference for cases not needing explicit type for the error
        # This branch might not be strictly necessary if all cases provide explicit_block_type or are fine with Conv2d default for 3-dim input_shape
        block_type = "Conv1d" if len(input_shape) < 3 else "Conv2d"

    with pytest.raises(AssertionError):
        EvolvableCNN(
            input_shape=input_shape,
            channel_size=channel_size,
            kernel_size=kernel_size,
            stride_size=stride_size,
            num_outputs=num_outputs,
            block_type=block_type,
            device=device,
        )


@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs",
    [([1, 16, 16], [3, 32], [3, 3], [2, 2], 10)],  # input_shape is (C, D, H, W)
)
def test_instantiation_for_multi_agents(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    device,
):
    evolvable_cnn = EvolvableCNN(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        block_type="Conv3d",
        sample_input=torch.randn(1, 1, *input_shape)
        .permute(0, 2, 1, 3, 4)
        .to(device),  # sample_input (B, C, D, H, W)
        device=device,
    )
    assert isinstance(evolvable_cnn, EvolvableCNN)


@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, sample_input, num_outputs",
    [
        ([1, 16, 16], [3, 32], [3, 3], [2, 2], "tensor", 10),
        ([1, 16, 16], [3, 32], [3, 3], [2, 2], None, 10),
    ],
)
def test_incorrect_instantiation_for_multi_agents(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    sample_input,
    num_outputs,
    device,
):
    if sample_input == "tensor":
        sample_input = torch.randn(1, *input_shape).to(device)

    with pytest.raises(AssertionError):
        EvolvableCNN(
            input_shape=input_shape,
            channel_size=channel_size,
            kernel_size=kernel_size,
            stride_size=stride_size,
            num_outputs=num_outputs,
            block_type="Conv3d",
            sample_input=sample_input,
            device=device,
        )


######### Test forward #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs, output_shape",
    [
        ([1, 16, 16], [32], [3], [1], 10, (1, 10)),
        ([1, 16, 16], [32], [(3, 3)], [(1, 1)], 10, (1, 10)),
        ([3, 128, 128], [8, 8, 8], [2, 2, 2], [2, 2, 2], 1, (1, 1)),
        ([1, 64], [32], [3], [1], 10, (1, 10)),  # Conv1D case
    ],
)
def test_forward(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    output_shape,
    device,
):
    block_type = "Conv1d" if len(input_shape) == 2 else "Conv2d"  # Infer block type
    evolvable_cnn = EvolvableCNN(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        block_type=block_type,
        device=device,
    )
    if block_type == "Conv1d":
        input_tensor = torch.randn(1, *input_shape).to(dtype=torch.float32)  # (B, C, L)
        input_array = np.random.randn(1, *input_shape)  # (B, C, L)
    else:  # Conv2d
        input_tensor = (
            torch.randn(input_shape).unsqueeze(0).to(dtype=torch.float32)
        )  # To add in a batch size dimension (B, C, H, W)
        input_array = np.expand_dims(np.random.randn(*input_shape), 0)  # (B, C, H, W)

    input_tensor = input_tensor.to(device)
    output = evolvable_cnn.forward(input_tensor)
    output_array = evolvable_cnn.forward(input_array)
    assert output.shape == output_shape
    assert output_array.shape == output_shape


@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, \
        num_outputs, output_shape",
    [([1, 16, 16], [3, 32], [3, 3], [2, 2], 10, (1, 10))],  # input_shape (C,H,W)
)
def test_forward_multi(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    output_shape,
    device,
):
    evolvable_cnn = EvolvableCNN(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        block_type="Conv3d",
        sample_input=torch.randn(1, 1, *input_shape)
        .permute(0, 2, 1, 3, 4)
        .to(device),  # sample_input (B,C,D,H,W)
        device=device,
    )
    input_tensor = torch.randn(1, *input_shape).to(device)  # input_tensor (B,C,D,H,W)
    with torch.no_grad():
        output = evolvable_cnn.forward(input_tensor)
    assert output.shape == output_shape


######### Test add_cnn_layer #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs",
    [
        ([1, 16, 16], [32], [3], [1], 10),
        ([3, 128, 128], [8, 8, 8], [2, 2, 2], [2, 2, 2], 1),
        ([1, 64], [32], [3], [1], 10),  # Conv1D case
    ],
)
def test_add_cnn_layer_simple(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    device,
):
    block_type = "Conv1d" if len(input_shape) == 2 else "Conv2d"
    evolvable_cnn = EvolvableCNN(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        block_type=block_type,
        device=device,
    )
    initial_channel_num = len(evolvable_cnn.channel_size)
    initial_net = evolvable_cnn.model
    initial_net_dict = dict(initial_net.named_parameters())
    evolvable_cnn.add_layer()
    new_net = evolvable_cnn.model
    if initial_channel_num < 6:
        assert len(evolvable_cnn.channel_size) == initial_channel_num + 1
        for key, param in new_net.named_parameters():
            if key in initial_net_dict.keys() and "linear_output" not in key:
                torch.testing.assert_close(param, initial_net_dict[key])
    else:
        assert len(evolvable_cnn.channel_size) == initial_channel_num


@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs",
    [
        ([1, 16, 16], [32, 32], [5, 5], [2, 2], 10),  # invalid output size
        (
            [1, 164, 164],
            [8, 8, 8, 8, 8, 8],
            [2, 2, 2, 2, 2, 2],
            [2, 2, 1, 1, 1, 1],
            10,
        ),  # exceeds max layer limit
        ([1, 8], [32, 32], [3, 3], [2, 2], 10),  # Conv1D, length too small
    ],
)
def test_add_cnn_layer_no_layer_added(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    device,
):
    block_type = "Conv1d" if len(input_shape) == 2 else "Conv2d"
    evolvable_cnn = EvolvableCNN(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        block_type=block_type,
        device=device,
    )
    evolvable_cnn.add_layer()
    assert len(channel_size) == len(evolvable_cnn.channel_size)


@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs",
    [
        ([3, 84, 84], [8, 8], [2, 2], [2, 2], 10),  # exceeds max-layer limit
        ([1, 128], [16, 16], [3, 3], [2, 2], 10),  # Conv1D case
    ],
)
def test_add_and_remove_multiple_cnn_layers(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    device,
):
    block_type = "Conv1d" if len(input_shape) == 2 else "Conv2d"
    evolvable_cnn = EvolvableCNN(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        block_type=block_type,
        device=device,
    )
    # Keep adding layers until we reach max or it is infeasible
    for _ in range(evolvable_cnn.max_hidden_layers):
        evolvable_cnn.add_layer()

    # Do a forward pass to ensure network parameter validity
    if block_type == "Conv1d":
        sample_data = torch.ones(1, input_shape[0], input_shape[1]).to(device)
    else:  # Conv2d/Conv3d (original was (1,3,84,84) - assuming input_shape[0] is num_channels)
        sample_data = torch.ones(1, input_shape[0], input_shape[1], input_shape[2]).to(
            device
        )

    output = evolvable_cnn(sample_data)
    assert output.squeeze().shape[0] == num_outputs
    assert len(evolvable_cnn.stride_size) == len(evolvable_cnn.channel_size)
    assert len(evolvable_cnn.mut_kernel_size) == len(evolvable_cnn.channel_size)

    # Remove as many layers as possible
    for _ in evolvable_cnn.channel_size:
        evolvable_cnn.remove_layer()

    assert len(evolvable_cnn.channel_size) == evolvable_cnn.min_hidden_layers

    # Do a forward pass to ensure network parameter validity
    output = evolvable_cnn(sample_data)
    assert output.squeeze().shape[0] == num_outputs
    assert len(evolvable_cnn.stride_size) == len(evolvable_cnn.channel_size)
    assert len(evolvable_cnn.mut_kernel_size) == len(evolvable_cnn.channel_size)


def test_add_cnn_layer_else_statement(device):
    evolvable_cnn = EvolvableCNN(
        input_shape=[1, 16, 16],
        channel_size=[32, 32],
        kernel_size=[3, 3],
        stride_size=[1, 1],
        num_outputs=4,
        max_hidden_layers=2,
        device=device,
    )
    original_num_hidden_layers = copy.deepcopy(evolvable_cnn.channel_size)
    evolvable_cnn.add_layer()
    assert len(original_num_hidden_layers) == len(evolvable_cnn.channel_size)


def test_add_cnn_layer_else_statement_conv1d(device):
    evolvable_cnn = EvolvableCNN(
        input_shape=[1, 64],  # (channels, length)
        channel_size=[32, 32],
        kernel_size=[3, 3],
        stride_size=[1, 1],
        num_outputs=4,
        max_hidden_layers=2,
        block_type="Conv1d",
        device=device,
    )
    original_num_hidden_layers = copy.deepcopy(evolvable_cnn.channel_size)
    evolvable_cnn.add_layer()
    assert len(original_num_hidden_layers) == len(evolvable_cnn.channel_size)


######### Test remove_cnn_layer #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs",
    [
        ([1, 16, 16], [32], [3], [1], 10),
        ([3, 128, 128], [8, 8, 8], [2, 2, 2], [2, 2, 2], 1),
        ([1, 64], [32, 32], [3, 3], [1, 1], 10),  # Conv1D case
    ],
)
def test_remove_cnn_layer(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    device,
):
    block_type = "Conv1d" if len(input_shape) == 2 else "Conv2d"
    evolvable_cnn = EvolvableCNN(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        block_type=block_type,
        device=device,
    )
    initial_channel_num = len(evolvable_cnn.channel_size)
    initial_net = evolvable_cnn.model
    initial_net_dict = dict(initial_net.named_parameters())
    evolvable_cnn.remove_layer()
    new_net = evolvable_cnn.model
    if initial_channel_num > 1:
        assert len(evolvable_cnn.channel_size) == initial_channel_num - 1
        for key, param in new_net.named_parameters():
            if (
                key in initial_net_dict.keys()
                and param.shape == initial_net_dict[key].shape
            ):
                torch.testing.assert_close(param, initial_net_dict[key])
    else:
        assert len(evolvable_cnn.channel_size) == initial_channel_num


######### Test add_cnn_channel #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs, layer_index",
    [
        ([1, 16, 16], [32], [3], [1], 10, 0),
        ([3, 128, 128], [8, 8, 8], [2, 2, 2], [2, 2, 2], 1, None),
        ([1, 64], [32], [3], [1], 10, 0),  # Conv1D case
    ],
)
def test_add_channels(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    layer_index,
    device,
):
    block_type = "Conv1d" if len(input_shape) == 2 else "Conv2d"
    evolvable_cnn = EvolvableCNN(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        block_type=block_type,
        device=device,
    )
    original_channel_size = copy.deepcopy(evolvable_cnn.channel_size)
    result = evolvable_cnn.add_channel(hidden_layer=layer_index)
    hidden_layer = result["hidden_layer"]
    numb_new_channels = result["numb_new_channels"]
    assert (
        evolvable_cnn.channel_size[hidden_layer]
        == original_channel_size[hidden_layer] + numb_new_channels
    )


######### Test remove_cnn_channel #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs, layer_index, numb_new_channels",
    [
        ([1, 16, 16], [256], [3], [1], 10, None, None),
        ([3, 128, 128], [8, 8, 8], [2, 2, 2], [2, 2, 2], 1, 0, 2),
        ([1, 64], [256], [3], [1], 10, None, None),  # Conv1D case
        ([1, 16, 16], [3], [3], [1], 10, 0, 8),  # exceeds min_channel_size
    ],
)
def test_remove_channels(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    layer_index,
    numb_new_channels,
    device,
):
    block_type = "Conv1d" if len(input_shape) == 2 else "Conv2d"
    evolvable_cnn = EvolvableCNN(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        min_channel_size=4,
        block_type=block_type,
        device=device,
    )
    original_channel_size = copy.deepcopy(evolvable_cnn.channel_size)
    result = evolvable_cnn.remove_channel(
        numb_new_channels=numb_new_channels, hidden_layer=layer_index
    )
    hidden_layer = result["hidden_layer"]
    numb_new_channels = result["numb_new_channels"]
    assert (
        evolvable_cnn.channel_size[hidden_layer]
        == original_channel_size[hidden_layer] - numb_new_channels
    )


######### Test change_cnn_kernel #########
def test_change_cnn_kernel(device):
    evolvable_cnn = EvolvableCNN(
        input_shape=[1, 16, 16],
        channel_size=[32, 32],
        kernel_size=[3, 3],
        stride_size=[1, 1],
        num_outputs=4,
        device=device,
    )
    # Change kernel size
    evolvable_cnn.change_kernel()

    while evolvable_cnn.mut_kernel_size == [(3, 3), (3, 3)]:
        evolvable_cnn.change_kernel()

    # Check if kernel size has changed
    assert evolvable_cnn.mut_kernel_size != [
        (3, 3),
        (3, 3),
    ], evolvable_cnn.mut_kernel_size


def test_change_kernel_size(device):
    evolvable_cnn = EvolvableCNN(
        input_shape=[1, 16, 16],
        channel_size=[32, 32],
        kernel_size=[3, 3],
        stride_size=[1, 1],
        num_outputs=4,
        device=device,
    )

    for _ in range(100):
        # Change kernel size and ensure we can make a valid forward pass
        evolvable_cnn.change_kernel()
        output = evolvable_cnn(torch.ones(1, 1, 16, 16).to(device))
        assert output.squeeze().shape[0] == 4  # (num actions)


def test_change_cnn_kernel_else_statement(device):
    evolvable_cnn = EvolvableCNN(
        input_shape=[1, 16, 16],
        channel_size=[32, 32],
        kernel_size=[3, 3],
        stride_size=[1, 1],
        num_outputs=4,
        device=device,
    )

    # Change kernel size
    evolvable_cnn.change_kernel()

    while evolvable_cnn.mut_kernel_size == [3, 3]:
        evolvable_cnn.change_kernel()

    # Check if kernel size has changed
    assert evolvable_cnn.mut_kernel_size != [3, 3]


def test_change_cnn_kernel_multi(device):
    evolvable_cnn = EvolvableCNN(
        input_shape=[1, 16, 16],  # (C,H,W)
        channel_size=[32, 32],
        kernel_size=[3, 3],
        stride_size=[1, 1],
        num_outputs=4,
        block_type="Conv3d",
        sample_input=torch.randn(1, 1, 1, 16, 16).to(device),  # (B,C,D,H,W)
        device=device,
    )

    # Change kernel size
    evolvable_cnn.change_kernel()

    while evolvable_cnn.mut_kernel_size.int_sizes == [3, 3]:
        evolvable_cnn.change_kernel()

    # Check if kernel size has changed
    assert evolvable_cnn.mut_kernel_size != [
        3,
        3,
    ], evolvable_cnn.mut_kernel_size


def test_change_cnn_kernel_multi_else_statement(device):
    evolvable_cnn = EvolvableCNN(
        input_shape=[1, 16, 16],  # (C,H,W)
        channel_size=[32],
        kernel_size=[3],
        stride_size=[1],
        block_type="Conv3d",
        sample_input=torch.randn(1, 1, 1, 16, 16).to(device),  # (B,C,D,H,W)
        num_outputs=4,
        device=device,
    )

    # Change kernel size
    while evolvable_cnn.mut_kernel_size.int_sizes == [3]:
        evolvable_cnn.change_kernel()

    # Check if kernel size has changed
    assert len(evolvable_cnn.mut_kernel_size) == 2


def test_change_cnn_kernel_conv1d(device):
    evolvable_cnn = EvolvableCNN(
        input_shape=[1, 64],  # (channels, length)
        channel_size=[32, 32],
        kernel_size=[3, 3],  # Will be converted to [(3,), (3,)]
        stride_size=[1, 1],
        num_outputs=4,
        block_type="Conv1d",
        device=device,
    )
    # Change kernel size
    evolvable_cnn.change_kernel()

    # Initial kernel sizes are [(3,), (3,)] due to MutableKernelSizes post_init for Conv1d
    initial_kernels = [(3,), (3,)]
    while evolvable_cnn.mut_kernel_size.sizes == initial_kernels:
        evolvable_cnn.change_kernel()

    # Check if kernel size has changed
    assert (
        evolvable_cnn.mut_kernel_size.sizes != initial_kernels
    ), evolvable_cnn.mut_kernel_size.sizes


def test_change_kernel_size_conv1d(device):
    evolvable_cnn = EvolvableCNN(
        input_shape=[1, 64],
        channel_size=[32, 32],
        kernel_size=[3, 3],
        stride_size=[1, 1],
        num_outputs=4,
        block_type="Conv1d",
        device=device,
    )

    for _ in range(100):
        # Change kernel size and ensure we can make a valid forward pass
        evolvable_cnn.change_kernel()
        output = evolvable_cnn(torch.ones(1, 1, 64).to(device))  # (B, C, L)
        assert output.squeeze().shape[0] == 4  # (num actions)


def test_change_cnn_kernel_else_statement_conv1d(device):
    evolvable_cnn = EvolvableCNN(
        input_shape=[1, 64],
        channel_size=[32, 32],
        kernel_size=[3, 3],  # Will be [(3,), (3,)]
        stride_size=[1, 1],
        num_outputs=4,
        block_type="Conv1d",
        device=device,
    )

    # Change kernel size
    evolvable_cnn.change_kernel()

    # mut_kernel_size.sizes will be list of tuples e.g. [(3,), (2,)]
    # mut_kernel_size.int_sizes will be [3,2]
    initial_kernels_int = [3, 3]

    while evolvable_cnn.mut_kernel_size.int_sizes == initial_kernels_int:
        evolvable_cnn.change_kernel()

    # Check if kernel size has changed
    assert evolvable_cnn.mut_kernel_size.int_sizes != initial_kernels_int


######### Test clone #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs",
    [
        ([1, 16, 16], [32], [3], [1], 10),
        ([3, 128, 128], [8, 8, 8], [2, 2, 2], [2, 2, 2], 1),
        ([1, 64], [32], [3], [1], 10),  # Conv1D case
    ],
)
def test_clone_instance(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    device,
):
    block_type = "Conv1d" if len(input_shape) == 2 else "Conv2d"
    evolvable_cnn = EvolvableCNN(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        block_type=block_type,
        device=device,
    )
    original_feature_net_dict = dict(evolvable_cnn.model.named_parameters())
    clone = evolvable_cnn.clone()
    clone_net = clone.model
    assert isinstance(clone, EvolvableCNN)
    assert_state_dicts_equal(clone.state_dict(), evolvable_cnn.state_dict())
    for key, param in clone_net.named_parameters():
        torch.testing.assert_close(param, original_feature_net_dict[key])
